#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import re
import tkinter as tk
import subprocess

from pathlib import Path, PureWindowsPath, PurePath
from threading import Thread
from tkinter import messagebox
from traceback import format_exc

from reclaimer.constants import inject_halo_constants

# before we do anything, we need to inject these constants so any definitions
# that are built that use them will have them in their descriptor entries.
inject_halo_constants()

from binilla.handler import Handler
from binilla.app_window import Binilla, default_hotkeys,\
     default_tag_window_hotkeys
from binilla.util import do_subprocess, open_in_default_program,\
    ProcController, is_main_frozen
from binilla.windows.about_window import AboutWindow
from binilla.windows.def_selector_window import DefSelectorWindow
from binilla.windows.filedialog import askopenfilename, askopenfilenames,\
    askdirectory, asksaveasfilename
from binilla.windows.tag_window import read_hotkey_string

from reclaimer.hek.handler import HaloHandler
from reclaimer.h3.handler import Halo3Handler
from reclaimer.os_v3_hek.handler import OsV3HaloHandler
from reclaimer.os_v4_hek.handler import OsV4HaloHandler
from reclaimer.misc.handler import MiscHaloLoader
from reclaimer.stubbs.handler import StubbsHandler
from supyr_struct.util import tagpath_to_fullpath, path_split,\
     path_replace, path_normalize, is_in_dir, is_path_empty

import mozzarilla

from mozzarilla import editor_constants as e_c
from mozzarilla.widgets.field_widget_picker import def_halo_widget_picker
from mozzarilla.widgets.directory_frame import DirectoryFrame
from mozzarilla.windows.tag_window import HaloTagWindow, HaloConfigWindow
from mozzarilla.windows.tools import \
     SearchAndReplaceWindow, SauceRemovalWindow, \
     BitmapSourceExtractorWindow, BitmapConverterWindow,\
     DependencyWindow, TagScannerWindow, DataExtractionWindow,\
     bitmap_from_dds, bitmap_from_multiple_dds, bitmap_from_bitmap_source, \
     AnimationsCompilerWindow, AnimationsCompressionWindow,\
     SoundCompilerWindow, ModelCompilerWindow,\
     physics_from_jms, hud_message_text_from_hmt, strings_from_txt
from mozzarilla.windows.tag_converters import ObjectConverter,\
     GbxmodelConverter, ModelConverter, ChicagoShaderConverter,\
     ModelAnimationsConverter, CollisionConverter, SbspConverter


default_hotkeys.update({
    '<F1>': "show_dependency_viewer",
    '<F2>': "show_tag_scanner",
    '<F3>': "show_search_and_replace",
    '<F4>': "show_data_extraction_window",

    '<F5>': "switch_tags_dir",
    '<F6>': "set_tags_dir",
    '<F7>': "add_tags_dir",
    #'<F8>': "???",

    '<F9>': "bitmap_from_dds",
    '<F10>': "bitmap_from_bitmap_source",
    '<F11>': "bitmap_from_multiple_dds",
    #'<F12>': "create_hek_pool_window",
    })

class Mozzarilla(Binilla):
    app_name = 'Mozzarilla'
    version = "%s.%s.%s" % mozzarilla.__version__
    log_filename = 'mozzarilla.log'
    debug = 0

    '''Directories/filepaths'''
    tags_dirs = ()

    _styles_dir  = Path(e_c.SETTINGS_DIR, "styles")
    _config_path = Path(e_c.SETTINGS_DIR, "mozzarilla.cfg")
    _last_data_load_dir = Path("")
    _jms_load_dir = Path("")
    _bitmap_load_dir = Path("")


    issue_tracker_url = mozzarilla.__website__ + "/issues"

    _mozzarilla_initialized = False
    guerilla_workspace_def  = None
    config_version = 3

    handler_classes = (
        HaloHandler,
        OsV3HaloHandler,
        OsV4HaloHandler,
        MiscHaloLoader,
        StubbsHandler,
        Halo3Handler,
        )

    handlers = ()

    handler_names = (
        "Halo 1",
        "Halo 1 OS v3",
        "Halo 1 OS v4",
        "Halo 1 Misc",
        "Stubbs the Zombie",
        "Halo 3"
        )

    # names of the handlers that MUST load tags from within their tags_dir
    tags_dir_relative = frozenset((
        "Halo 1",
        "Halo 1 OS v3",
        "Halo 1 OS v4",
        "Stubbs the Zombie",
        "Halo 3"
        ))

    about_module_names = (
        "arbytmap",
        "binilla",
        "mozzarilla",
        "reclaimer",
        "supyr_struct",
        "threadsafe_tkinter",
        )

    about_messages = ()
    color_names = e_c.mozz_color_names
    font_names = e_c.mozz_font_names

    _curr_handler_index  = 0
    _curr_tags_dir_index = 0

    widget_picker = def_halo_widget_picker
    config_window_class = HaloConfigWindow

    tool_windows = None

    window_panes = None
    directory_frame = None
    directory_frame_width = 200

    def __init__(self, *args, **kwargs):
        self.debug = kwargs.pop('debug', self.debug)

        # gotta give it a default handler or else the
        # config file will fail to be created as updating
        # the config requires using methods in the handler.
        kwargs['handler'] = MiscHaloLoader(debug=self.debug)
        try:
            with Path(e_c.MOZZLIB_DIR, "tad.gsm"[::-1]).open('r', -1, "037") as f:
                setattr(self, 'segassem_tuoba'[::-1], list(l for l in f))
        except Exception:
            pass

        tags_dir_fullpath = tagpath_to_fullpath(
            e_c.WORKING_DIR, "tags", folder=True)
        if tags_dir_fullpath is None:
            tags_dir_fullpath = e_c.WORKING_DIR.joinpath("tags")

        self.tags_dirs = [Path(tags_dir_fullpath)]
        self.handlers = list({} for i in range(len(self.handler_classes)))
        self.handler_names = list(self.handler_names)

        tk.Tk.__init__(self, *args, **{
            k: v for k, v in kwargs.items() if k in (
            "screenName", "baseName", "className", "useTk", "sync", "use"
            )})

        # NOTE: Do this import AFTER Tk interpreter is set up, otherwise
        # it will fail to get the names of the font families
        from mozzarilla.defs.config_def import config_def, mozz_config_version_def
        from mozzarilla.defs.v2_config_def import v2_config_def
        from mozzarilla.defs.guerilla_workspace_def import guerilla_workspace_def

        kwargs.update(
            config_def=config_def, config_version_def=mozz_config_version_def,
            config_defs={1: v2_config_def, 2: v2_config_def, 3: config_def})
        self.guerilla_workspace_def = guerilla_workspace_def

        Binilla.__init__(self, *args, **kwargs)

        self.app_bitmap_filepath = e_c.MOZZ_BITMAP_PATH

        if not e_c.IS_LNX:
            self.icon_filepath = e_c.MOZZ_ICON_PATH
            if self.icon_filepath:
                self.iconbitmap(str(self.icon_filepath))

        if is_path_empty(self.icon_filepath):
            print("Could not load window icon.")

        self.file_menu.insert_command("Exit", label="Load guerilla config",
                                      command=self.load_guerilla_config)
        self.file_menu.insert_separator("Exit")

        self.settings_menu.delete(0, "end")  # clear the menu
        self.settings_menu.add_command(label="Set tags directory",
                                       command=self.set_tags_dir)
        self.settings_menu.add_command(label="Add tags directory",
                                       command=self.add_tags_dir)
        self.settings_menu.add_command(label="Remove tags directory",
                                       command=self.remove_tags_dir)

        self.settings_menu.add_separator()
        self.settings_menu.add_command(
            label="Edit config", command=self.show_config_file)
        self.settings_menu.add_command(
            label="Open config directory", command=self.show_config_folder)
        self.settings_menu.add_separator()
        self.settings_menu.add_command(
            label="Load style", command=self.load_style)
        self.settings_menu.add_command(
            label="Save current style", command=self.make_style)
        self.settings_menu.add_command(
            label="Reset style", command=self.reset_style)

        # make the tools and tag set menus
        self.tools_menu = tk.Menu(self.main_menu, tearoff=0)
        self.compile_menu = tk.Menu(self.main_menu, tearoff=0)
        self.defs_menu = tk.Menu(self.main_menu, tearoff=0,
                                 postcommand=self.generate_defs_menu)
        self.converters_menu = tk.Menu(self.tools_menu, tearoff=0)

        self.main_menu.delete(0, "end")  # clear the menu
        self.main_menu.add_cascade(label="File",    menu=self.file_menu)
        self.main_menu.add_cascade(label="Settings", menu=self.settings_menu)
        self.main_menu.add_cascade(label="Tag Windows", menu=self.windows_menu)
        if self.debug_mode:
            self.main_menu.add_cascade(label="Debug", menu=self.debug_menu)
        self.main_menu.add_cascade(label="Tag set", menu=self.defs_menu)
        self.main_menu.add_cascade(label="Tools", menu=self.tools_menu)
        self.main_menu.add_cascade(label="Compile Tag", menu=self.compile_menu)
        self.main_menu.add_command(label="About", command=self.show_about_window)
        self.main_menu.add_command(label="Report Bug", command=self.open_issue_tracker)
        try:
            if e_c.IS_WIN and not is_main_frozen():
                import hek_pool
                self.main_menu.add_command(label="Launch Pool",
                                           command=self.create_hek_pool_window)
        except ImportError:
            pass

        self.tools_menu.add_command(
            label="Search and replace", command=self.show_search_and_replace)
        self.tools_menu.add_command(
            label="Scenario 'Open Sauce' remover",
            command=self.show_sauce_removal_window)
        self.tools_menu.add_separator()
        self.tools_menu.add_command(
            label="Bitmap converter",
            command=self.show_bitmap_converter_window)
        self.tools_menu.add_command(
            label="Bitmap source extractor",
            command=self.show_bitmap_source_extractor)
        self.tools_menu.add_separator()
        self.tools_menu.add_command(
            label="Model_animations decompressor",
            command=self.show_animations_compression_window)
        self.tools_menu.add_separator()
        self.tools_menu.add_command(
            label="Tags directory error locator", command=self.show_tag_scanner)
        self.tools_menu.add_command(
            label="Tag dependency viewer / zipper", command=self.show_dependency_viewer)
        self.tools_menu.add_command(
            label="Tag data extraction",
            command=self.show_data_extraction_window)
        self.tools_menu.add_separator()
        self.tools_menu.add_cascade(
            label="Tag converters", menu=self.converters_menu)

        self.converters_menu.add_command(
            label="scenario_structure_bsp  to  gbxmodel", command=self.show_sbsp_converter)
        self.converters_menu.add_command(
            label="model_collision_geometry  to  gbxmodel", command=self.show_collision_converter)
        self.converters_menu.add_command(
            label="model  to  gbxmodel", command=self.show_model_converter)
        self.converters_menu.add_separator()
        self.converters_menu.add_command(
            label="object  to  object", command=self.show_object_converter)
        self.converters_menu.add_separator()
        self.converters_menu.add_command(
            label="gbxmodel  to  model", command=self.show_gbxmodel_converter)
        self.converters_menu.add_separator()
        self.converters_menu.add_command(
            label="chicago_extended  to  chicago (shaders)", command=self.show_chicago_shader_converter)
        self.converters_menu.add_separator()
        self.converters_menu.add_command(
            label="model_animations_yelo  to  model_animations", command=self.show_animations_converter)

        self.compile_menu.add_command(
            label="Bitmap from dds texture(s)", command=self.bitmap_from_multiple_dds)
        self.compile_menu.add_command(
            label="Bitmap(s) from dds texture(s)", command=self.bitmap_from_dds)
        self.compile_menu.add_command(
            label="Bitmap(s) from bitmap source", command=self.bitmap_from_bitmap_source)
        self.compile_menu.add_separator()
        self.compile_menu.add_command(
            label="Sound from wav", command=self.show_sound_compiler_window)
        self.compile_menu.add_separator()
        self.compile_menu.add_command(
            label="Model_animations from jma", command=self.show_animations_compiler_window)
        self.compile_menu.add_command(
            label="Gbxmodel from jms", command=self.show_model_compiler_window)
        self.compile_menu.add_command(
            label="Physics from jms", command=self.physics_from_jms)
        self.compile_menu.add_separator()
        self.compile_menu.add_command(
            label="Hud_message_text from hmt", command=self.hud_message_text_from_hmt)
        self.compile_menu.add_command(
            label="String_list / unicode_string_list from txt", command=self.strings_from_txt)

        self.defs_menu.add_separator()

        self.select_defs(manual=False)
        self.tool_windows = {}

        self._mozzarilla_initialized = True

        self.make_window_panes()
        self.make_directory_frame(self.window_panes)
        self.make_io_text(self.window_panes)
        self.apply_style()

        if self.directory_frame is not None:
            self.directory_frame.highlight_tags_dir(self.tags_dir)

        try:
            if self.config_file.data.app_window.flags.load_last_workspace:
                self.load_last_workspace()
        except AttributeError:
            pass

        if self.config_made_anew:
            messagebox.showinfo(
                "Select your default tags directory",
                "Halo tags are all relative to their 'tags' root directory. "
                "After this prompt, you will be asked to select your tags "
                "directory. If you choose not to, a default one will be set.",
                parent=self)
            self.set_tags_dir()
            messagebox.showinfo(
                "About tags directories",
                "If you want to change/add/remove a tags directory, click the "
                "Settings menu and click either set, add, or remove. You may "
                "toggle through tags directories at any time with F5.",
                parent=self)

    @property
    def last_data_load_dir(self):
        return self._last_data_load_dir
    @last_data_load_dir.setter
    def last_data_load_dir(self, new_val):
        if not isinstance(new_val, Path):
            new_val = Path(new_val)
        self._last_data_load_dir = new_val

    @property
    def jms_load_dir(self):
        return self._jms_load_dir
    @jms_load_dir.setter
    def jms_load_dir(self, new_val):
        if not isinstance(new_val, Path):
            new_val = Path(new_val)
        self._jms_load_dir = new_val

    @property
    def bitmap_load_dir(self):
        return self._bitmap_load_dir
    @bitmap_load_dir.setter
    def bitmap_load_dir(self, new_val):
        if not isinstance(new_val, Path):
            new_val = Path(new_val)
        self._bitmap_load_dir = new_val

    @property
    def data_dir(self):
        tags_dir = self.tags_dir
        if tags_dir is None:
            return e_c.WORKING_DIR

        return path_replace(tags_dir, "tags", "data")

    @property
    def tags_dir(self):
        try:
            return Path(self.tags_dirs[self._curr_tags_dir_index])
        except IndexError:
            return None

    @tags_dir.setter
    def tags_dir(self, new_val):
        assert isinstance(new_val, (str, Path))
        self.tags_dirs[self._curr_tags_dir_index] = Path(new_val)

    def get_tags_dir_index(self, tags_dir):
        try:
            return self.tags_dirs.index(Path(tags_dir))
        except Exception:
            return None

    @property
    def handler_name(self):
        if self._curr_handler_index in range(len(self.handler_names)):
            return self.handler_names[self._curr_handler_index]
        return None

    def get_handler_index(self, handler):
        try:
            return self.handler_classes.index(type(handler))
        except Exception:
            return None

    def get_handler(self, index=None, tags_dir=None, create_if_not_exists=True):
        try:
            if index is None:
                index = self._curr_handler_index

            if not tags_dir:
                tags_dir = self.tags_dir

            if isinstance(index, str):
                index = self.handler_names.index(index)

            tags_dir = Path(tags_dir)

            if (self.handlers[index].get(tags_dir) is None and
                    create_if_not_exists):
                self.create_handlers(tags_dir, index)

            return self.handlers[index].get(tags_dir)
        except Exception:
            return None

    def create_handlers(self, tags_dir, handler_indices=()):
        tags_dir = Path(tags_dir)

        if isinstance(handler_indices, int):
            handler_indices = (handler_indices, )
        elif not handler_indices:
            handler_indices = range(len(self.handler_classes))

        for i in handler_indices:
            if isinstance(i, str):
                i = self.handler_names.index(i)

            if self.handlers[i].get(tags_dir) is not None:
                continue

            # TODO: Investigate.
            handler = self.handler_classes[i](debug=self.debug, case_sensitive=e_c.IS_LNX)
            handler.tagsdir = tags_dir
            self.handlers[i][tags_dir] = handler

    def set_active_handler(self, handler=None, index=None, tags_dir=None):
        if handler is not None:
            if not isinstance(handler, Handler):
                raise TypeError("Invalid type for handler argument. "
                                "Must be of type %s" % Handler)
            elif index is not None or tags_dir is not None:
                raise ValueError("Provide either a handler or a handler "
                                 "index and tags_dir, not all three.")

            index = self.get_handler_index(handler)
            tags_dir = handler.tagsdir
            if index is None:
                raise TypeError("Provided Handler is not recognized.")
        else:
            if not tags_dir:
                tags_dir = self.tags_dir
            elif not isinstance(tags_dir, (str, PurePath)):
                raise TypeError("Invalid type for tags_dir argument. Must be "
                                "of type %s, or %s, not %s" % (
                                    str, PurePath, type(tags_dir)))

            if index is None:
                index = self._curr_handler_index
            elif isinstance(index, str):
                index = self.handler_names.index(index)
            elif not isinstance(index, int):
                raise TypeError("Invalid type for index argument. Must be of "
                                "type %s or %s, not %s" % (str, int, type(index)))

            handler = self.get_handler(index, tags_dir)

        if None in (index, tags_dir):
            return

        tags_dir_index = self.get_tags_dir_index(tags_dir)
        self.handler = handler
        self._curr_handler_index = index
        if tags_dir_index is None:
            self.tags_dir = tags_dir
        else:
            self.switch_tags_dir(index=tags_dir_index, manual=False)

    def select_defs(self, menu_index=None, manual=True):
        if menu_index is None:
            menu_index = self._curr_handler_index
            # make sure the current handler index is valid
            if self.handler_name is None:
                menu_index = 0

        handler = self.get_handler(menu_index, create_if_not_exists=False)
        if not handler or handler is not self.handler:
            if manual:
                print("Changing tag set to %s" % self.handler_names[menu_index])
                # The first time this function is called is before initialization is
                # finished. This attribute will not exist yet by then.
                if hasattr(self, "io_text"):
                    self.io_text.update_idletasks()

            self.set_active_handler(index=menu_index)
            try:
                self.config_file.data.mozzarilla.selected_handler.data = menu_index
            except AttributeError:
                pass

            if manual:
                print("    Finished")

            # Update window title to reflect tag set that we're using.
        self.update_title()

    def generate_defs_menu(self):
        self.defs_menu.delete(0, "end")  # clear the menu
        for i in range(len(self.handler_names)):
            label = self.handler_names[i]
            if i == self._curr_handler_index:
                label += u' \u2713'
            self.defs_menu.add_command(label=label, command=lambda i=i:
                                       self.select_defs(i, manual=True))

    def add_tags_dir(self, e=None, tags_dir=None, manual=True):
        if tags_dir is None:
            tags_dir = askdirectory(initialdir=self.tags_dir, parent=self,
                                    title="Select the tags directory to add")

        if not tags_dir:
            return

        tags_dir = Path(tags_dir)
        if self.get_tags_dir_index(tags_dir) is not None:
            if manual:
                print("That tags directory already exists.")
            return

        self.tags_dirs.append(tags_dir)
        self.switch_tags_dir(index=len(self.tags_dirs) - 1, manual=False)

        if self.directory_frame is not None:
            self.directory_frame.add_root_dir(tags_dir)

        if manual:
            self.last_load_dir = tags_dir
            print("Tags directory is currently:\n    %s\n" % self.tags_dir)

    def remove_tags_dir(self, e=None, index=None, manual=True):
        dirs_count = len(self.tags_dirs)
        # need at least 2 tags dirs to delete one manually
        if dirs_count < 2 and manual:
            print("You need more than one tags directory before you're "
                  "allowed to remove one.")
            return

        if index is None:
            index = self._curr_tags_dir_index

        new_index = self._curr_tags_dir_index
        if index <= new_index:
            new_index = max(0, new_index - 1)

        tags_dir = self.tags_dirs[index]
        del self.tags_dirs[index]
        if self.directory_frame is not None:
            self.directory_frame.del_root_dir(tags_dir)

        self.switch_tags_dir(index=new_index, manual=False)

        if manual:
            self.last_load_dir = self.tags_dir
            print("Tags directory is currently:\n    %s\n" % self.tags_dir)

    def set_tags_dir(self, e=None, tags_dir=None, manual=True):
        if tags_dir is None:
            tags_dir = askdirectory(initialdir=self.tags_dir, parent=self,
                                    title="Select the tags directory")

        if not tags_dir:
            return

        tags_dir = Path(tags_dir)

        if tags_dir in self.tags_dirs:
            print("That tags directory already exists.")
            return

        if self.directory_frame is not None:
            self.directory_frame.set_root_dir(tags_dir)
            self.directory_frame.highlight_tags_dir(self.tags_dir)

        self.tags_dir = tags_dir
        self.set_active_handler()

        if manual:
            self.last_load_dir = self.tags_dir
            print("Tags directory is currently:\n    %s\n" % self.tags_dir)

    def switch_tags_dir(self, e=None, index=None, manual=True):
        if index is None:
            index = (self._curr_tags_dir_index + 1) % len(self.tags_dirs)
        if self._curr_tags_dir_index == index:
            return

        self._curr_tags_dir_index = index
        self.set_active_handler()

        if self.directory_frame is not None:
            self.directory_frame.highlight_tags_dir(self.tags_dir)

        if manual:
            self.last_load_dir = self.tags_dir
            print("Tags directory is currently:\n    %s\n" % self.tags_dir)

    def apply_config(self, e=None):
        Binilla.apply_config(self)
        config_data = self.config_file.data
        mozz = config_data.mozzarilla
        self._curr_handler_index = mozz.selected_handler.data
        tags_dirs = mozz.tags_dirs
        load_dirs = mozz.load_dirs

        try:
            self.select_defs()
        except Exception:
            print(format_exc())

        for i in range(len(self.tags_dirs)):
            self.remove_tags_dir(i, manual=False)

        self._curr_tags_dir_index = 0
        for tags_dir in tags_dirs:
            self.add_tags_dir(tags_dir=tags_dir.path, manual=False)

        backup_dir_basename = config_data.tag_backup.folder_basename
        for handler_set in self.handlers:
            for handler in handler_set.values():
                handler.backup_dir_basename = backup_dir_basename

        self.switch_tags_dir(
            index=min(mozz.last_tags_dir, len(self.tags_dirs)), manual=False)

        for s in ("last_data_load_dir", "jms_load_dir",
                  "bitmap_load_dir")[:len(load_dirs)]:
            try: setattr(self, s, Path(load_dirs[s].path))
            except IndexError: pass

        if not self.tags_dir:
            self.tags_dir = tagpath_to_fullpath(e_c.WORKING_DIR, "tags", folder=True)

    def record_open_tags(self):
        try:
            open_tags = self.config_file.data.mozzarilla.open_mozz_tags
            tags_dirs = self.config_file.data.mozzarilla.tags_dirs
            del open_tags[:]
        except Exception:
            print(format_exc())
            return

        for wid in sorted(self.tag_windows):
            try:
                w = self.tag_windows[wid]
                tag, handler = w.tag, w.handler

                # dont store tags that arent from the current handler
                if tag in (self.config_file, None):
                    continue

                tags_dir_index = self.get_tags_dir_index(handler.tagsdir)
                handler_index = self.get_handler_index(handler)
                if None in (tags_dir_index, handler_index):
                    continue

                open_tags.append()
                open_tag = open_tags[-1]
                header = open_tag.header

                if w.state() == 'withdrawn':
                    header.flags.minimized = True

                pos_x, pos_y = w.winfo_x(), w.winfo_y()
                width, height = w.geometry().split('+')[0].split('x')[:2]

                header.offset_x, header.offset_y = pos_x, pos_y
                header.width, header.height = int(width), int(height)
                header.tags_dir_index = tags_dir_index
                header.handler_index = handler_index

                open_tag.def_id, open_tag.path = tag.def_id, str(tag.filepath)
            except Exception:
                print(format_exc())

    def load_last_workspace(self):
        if not self._mozzarilla_initialized:
            return

        try:
            mozz = self.config_file.data.mozzarilla
            open_tags = mozz.open_mozz_tags
            tags_dirs = [b.path for b in mozz.tags_dirs]
        except Exception:
            print(format_exc())
            return

        print("Loading last workspace...")
        self.io_text.update_idletasks()
        for open_tag in open_tags:
            try:
                header = open_tag.header
                self.set_active_handler(
                    index=header.handler_index,
                    tags_dir=tags_dirs[header.tags_dir_index])

                windows = self.load_tags(filepaths=open_tag.path,
                                         def_id=open_tag.def_id)
                if not windows:
                    continue

                if header.flags.minimized:
                    windows[0].withdraw()
                    self.selected_tag = None

                windows[0].geometry("%sx%s+%s+%s" % (
                    header.width, header.height,
                    header.offset_x, header.offset_y))
            except Exception:
                print(format_exc())

        print("    Finished")

    def load_guerilla_config(self):
        if self.handler_name not in self.tags_dir_relative:
            print("Change the current tag set.")
            return

        fp = askopenfilename(initialdir=str(self.last_load_dir), parent=self,
                             title="Select the tag to load",
                             filetypes=(('Guerilla config', '*.cfg'),
                                        ('All', '*')))

        if not fp:
            return

        fp = Path(fp)

        self.last_load_dir = fp.parent
        tags_dir = tagpath_to_fullpath(fp.parent, "tags", folder=True)
        if tags_dir is None:
            print("Specified guerilla.cfg has no corresponding tags directory.")
            return

        tags_dir = Path(tags_dir)
        workspace = self.guerilla_workspace_def.build(filepath=fp)
        if self.get_tags_dir_index(tags_dir) is None:
            print("Adding tags directory:\n    %s" % tags_dir)
            self.add_tags_dir(tags_dir=tags_dir)

        self.set_active_handler(tags_dir=tags_dir)

        for tag in workspace.data.tags:
            if not(tag.is_valid_tag and not is_path_empty(tag.filepath)):
                continue

            windows = self.load_tags(tag.filepath)
            if not windows:
                continue

            w = windows[0]

            tl_corner = tag.window_header.t_l_corner
            br_corner = tag.window_header.b_r_corner

            self.place_window_relative(w, tl_corner.x, tl_corner.y)
            w.geometry("%sx%s" % (br_corner.x - tl_corner.x,
                                  br_corner.y - tl_corner.y))

    def load_tags(self, filepaths=None, def_id=None):
        tags_dir = self.tags_dir
        # if there is not tags directory, this can be loaded normally
        if tags_dir is None:
            return Binilla.load_tags(self, filepaths, def_id)

        if filepaths is None:
            filetypes = [('All', '*')]
            defs = self.handler.defs
            for id in sorted(defs.keys()):
                filetypes.append((id, defs[id].ext))
            filepaths = askopenfilenames(initialdir=str(self.last_load_dir),
                                         filetypes=filetypes, parent=self,
                                         title="Select the tag to load")

            if not filepaths:
                return ()
            elif isinstance(filepaths, str) and filepaths.startswith('{'):
                # account for a stupid bug with certain versions of windows
                filepaths = re.split("\}\W\{", filepaths[1:-1])

        if isinstance(filepaths, (str, PurePath)):
            filepaths = (filepaths, )

        if not filepaths:
            return ()

        filepaths = list(Path(fp) for fp in filepaths)

        # make sure all the chosen tag paths are relative
        # to the current tags directory if they must be
        last_load_dir = self.last_load_dir
        if self.handler_name in self.tags_dir_relative:
            for path in filepaths:
                if is_path_empty(path):
                    # path is empty, so making a new tag
                    continue
                elif is_in_dir(path, tags_dir):
                    last_load_dir = path.parent
                    continue
                else:
                    print(
                        "Not loading tag:\n"
                        "    %s\n"
                        "\n"
                        "Reason: Not located in current tags folder:\n"
                        "    %s"
                        % (path, tags_dir))
                    return ()

        windows = Binilla.load_tags(self, filepaths, def_id)
        self.last_load_dir = last_load_dir

        if not windows:
            print("You might need to change the tag set to load these tag(s).")
            return ()

        return windows

    def load_tag_as(self, e=None):
        '''Prompts the user for a tag to load and loads it.'''
        if self.def_selector_window:
            return

        filetypes = [('All', '*')]
        defs = self.handler.defs
        for def_id in sorted(defs.keys()):
            filetypes.append((def_id, defs[def_id].ext))

        fp = askopenfilename(initialdir=str(self.last_load_dir),
                             filetypes=filetypes, parent=self,
                             title="Select the tag to load")

        if not fp:
            return

        self.last_load_dir = Path(fp).parent
        dsw = DefSelectorWindow(
            self, title="Which tag is this", action=lambda def_id:
            self.load_tags(filepaths=fp, def_id=def_id))
        self.def_selector_window = dsw
        self.place_window_relative(self.def_selector_window, 30, 50)

    def new_tag(self, e=None):
        if self.def_selector_window:
            return

        dsw = DefSelectorWindow(
            self, title="Select a tag to create", action=lambda def_id:
            self.load_tags(filepaths='', def_id=def_id))
        self.def_selector_window = dsw
        self.place_window_relative(self.def_selector_window, 30, 50)

    def save_tag(self, tag=None):
        if tag is None:
            if self.selected_tag is None:
                print("Cannot save(no tag is selected).")
                return
            tag = self.selected_tag

        if tag is self.config_file:
            return self.save_config()

        # change the tags filepath to be relative to the current tags directory
        if hasattr(tag, "rel_filepath"):
            full_filepath = tagpath_to_fullpath(tag.tags_dir, tag.rel_filepath)
            if full_filepath:
                tag.filepath = full_filepath
            else:
                tag.filepath = Path(tag.tags_dir, tag.rel_filepath)

        return Binilla.save_tag(self, tag)

    def save_tag_as(self, tag=None, filepath=None):
        if tag is None:
            if self.selected_tag is None:
                print("Cannot save(no tag is selected).")
                return
            tag = self.selected_tag

        if not hasattr(tag, "serialize"):
            return

        if filepath is None:
            ext = tag.ext
            filepath = asksaveasfilename(
                initialdir=Path(tag.filepath).parent, parent=self,
                defaultextension=ext, title="Save tag as...",
                filetypes=[(ext[1:], "*" + ext), ('All', '*')])
        else:
            filepath = tag.filepath

        if is_path_empty(filepath):
            return

        # make sure to flush any changes made using widgets to the tag
        w = self.get_tag_window_by_tag(tag)

        # make sure the filepath is sanitized
        filepath = Path(path_normalize(filepath))

        handler = tag.handler
        tags_dir = self.tags_dir
        tagsdir_rel = handler.tagsdir_relative

        try:
            self.last_load_dir = filepath.parent
            if tagsdir_rel and not is_in_dir(filepath, tag.tags_dir):
                messagebox.showerror(
                    "Saving outside tags directory", ("Cannot save:\n\n"
                    "    %s\n\noutside the tags directory:\n\n    %s\n\n"
                    "Change the tags directory back to save this tag.") %
                    (filepath, tag.tags_dir), parent=self.focus_get())
                return

            self.add_tag(tag, filepath)
            w.save(temp=False)
        except PermissionError:
            print("This program does not have permission to save to this folder.\n"
                  "Could not save: %s" % filepath)
        except Exception:
            print(format_exc())
            raise IOError("Could not save: %s" % filepath)

        self.update_tag_window_title(w)
        return tag

    def save_all(self, e=None):
        '''
        Saves all currently loaded tags to their files.
        '''
        for handler_set in self.handlers:
            for handler in handler_set.values():
                if not getattr(handler, "tags", None):
                    continue

                tags = handler.tags
                for def_id in tags:
                    tag_coll = tags[def_id]
                    for tag_path in tag_coll:
                        try:
                            self.save_tag(tag_coll[tag_path])
                        except Exception:
                            print(format_exc())
                            print("Exception occurred while trying to save '%s'" %
                                  tag_path)

    def make_config(self, filepath=None):
        if filepath is None:
            filepath = self.config_path

        filepath = Path(filepath)

        # Make directory if it doesn't exist
        filepath.parent.mkdir(parents=True, exist_ok=True)

        # create the config file from scratch
        self.config_file = self.config_def.build()
        self.config_file.filepath = filepath

        data = self.config_file.data

        # make sure these have as many entries as they're supposed to
        for block in (data.directory_paths, data.appearance.colors,
                      data.appearance.depths):
            block.extend(len(block.NAME_MAP))

        tags_dirs = data.mozzarilla.tags_dirs
        for tags_dir in self.tags_dirs:
            tags_dirs.append()
            tags_dirs[-1].path = str(tags_dir)

        self.update_config()

        c_hotkeys = data.all_hotkeys.hotkeys
        c_tag_window_hotkeys = data.all_hotkeys.tag_window_hotkeys

        for k_set, b in ((default_hotkeys, c_hotkeys),
                         (default_tag_window_hotkeys, c_tag_window_hotkeys)):
            default_keys = k_set
            hotkeys = b
            for combo, method in k_set.items():
                hotkeys.append()
                keys = hotkeys[-1].combo

                modifier, key = read_hotkey_string(combo)
                keys.modifier.set_to(modifier)
                keys.key.set_to(key)

                hotkeys[-1].method.set_to(method)

    def update_config(self, config_file=None):
        if config_file is None:
            config_file = self.config_file
        Binilla.update_config(self, config_file)

        config_data = config_file.data
        mozz = config_data.mozzarilla
        tags_dirs = mozz.tags_dirs
        load_dirs = mozz.load_dirs

        mozz.selected_handler.data = self._curr_handler_index
        mozz.last_tags_dir = self._curr_tags_dir_index

        del tags_dirs[:]
        for tags_dir in self.tags_dirs:
            tags_dirs.append()
            tags_dirs[-1].path = str(path_normalize(tags_dir))

        if len(load_dirs.NAME_MAP) > len(load_dirs):
            load_dirs.extend(len(load_dirs.NAME_MAP) - len(load_dirs))

        for s in ("last_data_load_dir", "jms_load_dir", "bitmap_load_dir"):
            try: load_dirs[s].path = str(getattr(self, s))
            except IndexError: pass

        if mozz.flags.show_hierarchy_window and mozz.flags.show_console_window:
            try:
                # idk if this value can ever be negative, so i'm using abs
                mozz.sash_position = abs(self.window_panes.sash_coord(0)[0])
            except Exception:
                pass

    def update_tag_window_title(self, window):
        if not hasattr(window, 'tag'):
            return

        tag = window.tag
        if tag is self.config_file:
            window.update_title('%s %s config' % (self.app_name, self.version))

        if not hasattr(tag, 'tags_dir'):
            return

        tags_dir = tag.tags_dir

        try:
            if tag is self.config_file or not tags_dir:
                return
            if self.handler_name not in self.tags_dir_relative:
                window.update_title()
                return

            try:
                show_full = self.config_file.data.mozzarilla.\
                            flags.show_full_tags_directory
            except Exception:
                show_full = False

            tags_dir_str = tags_dir
            if not show_full:
                tags_dir_str = Path(tags_dir).name
                if tags_dir_str.lower() == 'tags':
                    tags_dir_str = Path(tags_dir).parent.name

            handler_i = self.get_handler_index(window.handler)

            title = "[%s][%s][%s]" % (
                self.handler_names[handler_i], tags_dir_str, tag.rel_filepath)
        except Exception:
            print(format_exc())
            title = window.title()

        window.update_title(title)

    def apply_style(self, seen=None):
        if not self._initialized:
            return

        with self.style_change_lock:
            try:
                Binilla.apply_style(self, seen)

                mozz = self.config_file.data.mozzarilla
                show_hierarchy = mozz.flags.show_hierarchy_window
                show_console = mozz.flags.show_console_window
                sash_pos = mozz.sash_position
                self.window_panes.forget(self.directory_frame)
                self.window_panes.forget(self.io_frame)

                if show_hierarchy:
                    self.directory_frame.pack(fill='both', expand=True)
                    self.window_panes.add(self.directory_frame)

                if show_console:
                    self.io_frame.pack(fill='both', expand=True)
                    self.window_panes.add(self.io_frame)

                # if both window panes are shown, we need to position the sash
                if show_hierarchy and show_console and sash_pos != 0:
                    try:
                        self.update_idletasks()
                        self.window_panes.sash_place(0, sash_pos, 1)
                    except Exception:
                        pass
            except AttributeError: print(format_exc())
            except Exception: print(format_exc())
            except Exception: print(format_exc())

    def make_tag_window(self, tag, *, focus=True, window_cls=None,
                        is_new_tag=False):
        if window_cls is None:
            window_cls = HaloTagWindow
        w = Binilla.make_tag_window(self, tag, focus=focus,
                                    window_cls=window_cls,
                                    is_new_tag=is_new_tag)
        self.update_tag_window_title(w)
        return w

    def make_window_panes(self):
        self.window_panes = tk.PanedWindow(
            self.root_frame, sashrelief='raised', sashwidth=8,
            bd=self.frame_depth, bg=self.frame_bg_color)
        self.window_panes.pack(anchor='nw', fill='both', expand=True)

    def make_io_text(self, master=None):
        if not self._initialized:
            return
        if master is None:
            master = self.root_frame
        Binilla.make_io_text(self, master)

    def make_directory_frame(self, master=None):
        if not self._initialized:
            return
        if master is None:
            master = self.root_frame
        self.directory_frame = DirectoryFrame(self)
        self.directory_frame.pack(expand=True, fill='both')

    def show_tool_window(self, window_name, window_class,
                         needs_tag_refs=False, **kw):
        w = self.tool_windows.get(window_name)
        if w is not None:
            try: del self.tool_windows[window_name]
            except Exception: pass
            try: w.destroy()
            except Exception: pass
            return

        if needs_tag_refs and not hasattr(self.handler, 'tag_ref_cache'):
            print("Change the current tag set.")
            return

        self.tool_windows[window_name] = w = window_class(self, **kw)
        w.window_name = window_name
        self.place_window_relative(w, 30, 50); w.focus_set()

    def show_chicago_shader_converter(self, e=None):
        self.show_tool_window("chicago_shader_converter", ChicagoShaderConverter)
    def show_object_converter(self, e=None):
        self.show_tool_window("object_converter_window", ObjectConverter)
    def show_model_converter(self, e=None):
        self.show_tool_window("model_converter_window", ModelConverter)
    def show_gbxmodel_converter(self, e=None):
        self.show_tool_window("gbxmodel_converter_window", GbxmodelConverter)
    def show_collision_converter(self, e=None):
        self.show_tool_window("collision_converter_window", CollisionConverter)
    def show_animations_converter(self, e=None):
        self.show_tool_window("animations_converter_window", ModelAnimationsConverter)
    def show_sbsp_converter(self, e=None):
        self.show_tool_window("sbsp_converter_window", SbspConverter)

    def show_dependency_viewer(self, e=None):
        self.show_tool_window("dependency_window", DependencyWindow, True)
    def show_tag_scanner(self, e=None):
        self.show_tool_window("tag_scanner_window", TagScannerWindow, True)

    def show_bitmap_converter_window(self, e=None):
        self.show_tool_window("bitmap_converter_window", BitmapConverterWindow)
    def show_bitmap_source_extractor(self, e=None):
        self.show_tool_window("bitmap_source_extractor_window", BitmapSourceExtractorWindow)

    def show_data_extraction_window(self, e=None):
        self.show_tool_window("data_extraction_window", DataExtractionWindow, True)

    def show_animations_compression_window(self, e=None):
        self.show_tool_window("animations_compression_window", AnimationsCompressionWindow, False)
    def show_animations_compiler_window(self, e=None):
        self.show_tool_window("animations_compiler_window", AnimationsCompilerWindow, True)
    def show_model_compiler_window(self, e=None):
        self.show_tool_window("model_compiler_window", ModelCompilerWindow, True)
    def show_sound_compiler_window(self, e=None):
        self.show_tool_window("sound_compiler_window", SoundCompilerWindow, False)

    def show_search_and_replace(self, e=None):
        self.show_tool_window("s_and_r_window", SearchAndReplaceWindow)

    def show_sauce_removal_window(self, e=None):
        self.show_tool_window("sauce_removal_window", SauceRemovalWindow)

    def create_hek_pool_window(self, e=None):
        try:
            launcher = Thread(
                target=do_subprocess, daemon=True, args=("pythonw", ),
                kwargs=dict(exec_args=("-m", "hek_pool.run"),
                            proc_controller=ProcController(abandon=True)))
            launcher.start()
        except Exception:
            print("Could not open HEK Pool")

    def bitmap_from_dds(self, e=None):
        bitmap_from_dds(self)

    def bitmap_from_bitmap_source(self, e=None):
        bitmap_from_bitmap_source(self)

    def bitmap_from_multiple_dds(self, e=None):
        bitmap_from_multiple_dds(self)

    def physics_from_jms(self, e=None):
        physics_from_jms(self)

    def hud_message_text_from_hmt(self, e=None):
        hud_message_text_from_hmt(self)

    def strings_from_txt(self, e=None):
        strings_from_txt(self)

    def upgrade_config_version(self, filepath):
        old_version = self.config_version_def.build(filepath=str(filepath)).data.version
        if old_version in (1, 2):
            new_config = mozzarilla.defs.upgrade_config.upgrade_v2_to_v3(
                self.config_defs[2].build(filepath=str(filepath)),
                self.config_defs[3].build())
        else:
            raise ValueError("Config header version is not valid")

        return new_config

    def show_config_folder(self, **kw):
        open_in_default_program(e_c.SETTINGS_DIR)

    def update_title(self):
        self.title('%s v%s [%s]' % (self.app_name, self.version, self.handler_name))

    def show_about_window(self):
        w = getattr(self, "about_window", None)
        if w is not None:
            try: w.destroy()
            except Exception: pass
            self.about_window = None

        if not hasattr(AboutWindow, "orig_pressed"):
            AboutWindow.orig_pressed = AboutWindow._pressed
            AboutWindow._pressed = self.some_func

        self.about_window = AboutWindow(
            self, module_names=self.about_module_names,
            iconbitmap=self.icon_filepath, appbitmap=self.app_bitmap_filepath,
            app_name=self.app_name, messages=self.about_messages)
        self.place_window_relative(self.about_window, 30, 50)

    def some_func(self):
        val = "734531alli6dgrwretsaM"
        self.app_name = val[::-1][:6]+val[::-1][11:11+4]
        self.version = val[::-1][-6:len(val)-4]+val[::-1][-2:]
        self.update_title()
        AboutWindow.orig_pressed(self.about_window)



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from supyr_struct.defs.tag_def import TagDef
from supyr_struct.field_types import *

__all__ = (
    "get", "guerilla_workspace_def",
    )


def reflexives_size(parent=None, new_value=None, **kwargs):
    if parent is None:
        raise KeyError()
    if new_value is None:
        return parent.reflexive_count * 4

    parent.reflexive_count = new_value // 4


def has_next_tag(rawdata=None, **kwargs):
    '''Returns whether or not there is another block in the stream.'''
    try:
        offset = kwargs.get('offset')
        try:
            offset += kwargs.get('root_offset')
        except Exception:
            pass
        return rawdata.peek(4, offset) == b'\x01\x00\x00\x00'
    except AttributeError:
        return False

reflexive_counts = {
    "actv": 1, "tagc": 1, "mgs2": 1, "lens": 1,
    "elec": 2,
    "bitm": 3, "sky ": 3, "phys": 3,
    "obje": 6, "eqip": 6, "garb": 6, "scen": 6,
    "plac": 6, "mach": 6, "lifi": 6, "ctrl": 6,
    "proj": 7,
    "unit": 8,
    "mode": 12, "mod2": 12,
    "antr": 22,
    "coll": 15, "bipd": 15,
    "matg": 19,
    "sbsp": 53,
    "scnr": 61,
    # This is incomplete
    }

window_header = Struct("window_header",
    UInt32("struct_size", DEFAULT=44),
    UInt32("unknown1"),
    UInt32("unknown2", DEFAULT=1),
    # These raw bytes seem to be some sort of window coordinates, but idc
    BytesRaw("unknown3", DEFAULT=b'\xff'*16, SIZE=16),

    QStruct("t_l_corner", SInt32("x"), SInt32("y"), ORIENT="h"),
    QStruct("b_r_corner", SInt32("x"), SInt32("y"), ORIENT="h"),
    SIZE=44
    )

open_halo_tag = Container("open_tag",
    UInt32("is_valid_tag", DEFAULT=1),
    window_header,
    UInt8("filepath_len"),
    StrRawAscii("filepath", SIZE='.filepath_len'),
    Pad(8),
    UInt16("reflexive_count"),

    # this seems to contain the indices that the
    # reflexives were on when the tag was last open
    SInt32Array("reflexive_indices", SIZE=reflexives_size),
    )

guerilla_workspace_def = TagDef("guerilla_workspace",
    window_header,
    WhileArray("tags",
        SUB_STRUCT=open_halo_tag,
        CASE=has_next_tag
        ),
    UInt32("eof_marker"),

    ENDIAN='<', ext=".cfg"
    )


def get(): return guerilla_workspace_def



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os

from pathlib import Path
from traceback import format_exc

from reclaimer.strings.strings_compilation import compile_unicode_string_list,\
     compile_string_list
from supyr_struct.util import is_path_empty
from binilla.windows.filedialog import askopenfilename


def strings_from_txt(app, fp=None):
    load_dir = app.last_data_load_dir
    tags_dir = app.tags_dir
    data_dir = app.data_dir
    if is_path_empty(tags_dir):
        tags_dir = Path("")
    if is_path_empty(data_dir):
        data_dir = Path("")

    if is_path_empty(load_dir):
        load_dir = data_dir

    if is_path_empty(fp):
        fp = askopenfilename(
            initialdir=load_dir, parent=app,
            filetypes=(("strings list", "*.txt"), ("All", "*")),
            title="Select hmt file to turn into a (unicode_)string_list tag")

    fp = Path(fp)
    if is_path_empty(fp):
        return

    try:
        app.last_data_load_dir = fp.parent
        tag_ext = "unicode_string_list"
        tag_cls = "ustr"

        with fp.open("rb") as f:
            data = f.read(2)

        if data[0] == 255 and data[1] == 254:
            encoding = "utf-16-le"
        elif data[1] == 254 and data[0] == 255:
            encoding = "utf-16-be"
        else:
            encoding = "latin-1"
            tag_ext = "string_list"
            tag_cls = "str#"

        print("Creating %s from this txt file:" % tag_ext)
        print("    %s" % fp)

        with fp.open("r", encoding=encoding) as f:
            string_data = f.read()
            if "utf-16" in encoding:
                string_data = string_data.lstrip('\ufeff').lstrip('\ufffe')

    except Exception:
        print(format_exc())
        print("    Could not parse file.")
        return

    try:
        rel_filepath = fp.relative_to(data_dir)
    except ValueError:
        rel_filepath = Path("strings")

    rel_filepath = rel_filepath.with_suffix("." + tag_ext)

    tag_path = Path("")
    if not is_path_empty(rel_filepath):
        tag_path = tags_dir.joinpath(rel_filepath)

    # make the tag window
    window = app.load_tags(
        filepaths=(tag_path, ) if tag_path.is_file() else "",
        def_id=tag_cls)
    if not window:
        return

    window = window[0]
    window.is_new_tag = True
    window.tag.filepath = tag_path
    window.tag.rel_filepath = rel_filepath

    if 'utf-16' in encoding:
        compile_unicode_string_list(window.tag, string_data)
    else:
        compile_string_list(window.tag, string_data)

    # reload the window to display the newly entered info
    window.reload()
    app.update_tag_window_title(window)
    if not tag_path.is_file():
        # save the tag if it doesnt already exist
        app.save_tag()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os

from pathlib import Path
from struct import unpack
from traceback import format_exc

from reclaimer.bitmaps.bitmap_decompilation import extract_bitmap_tiff_data
from reclaimer.bitmaps.bitmap_compilation import add_bitmap_to_bitmap_tag,\
     compile_bitmap_from_dds_files
from supyr_struct.util import is_path_empty
from binilla.windows.filedialog import askopenfilenames


def bitmap_from_dds(app, fps=()):
    load_dir = app.bitmap_load_dir
    data_dir = app.data_dir
    if is_path_empty(data_dir):
        data_dir = Path("")

    if is_path_empty(load_dir):
        load_dir = data_dir

    if not fps:
        fps = askopenfilenames(
            initialdir=load_dir, parent=app,
            filetypes=(("DDS image", "*.dds"), ("All", "*")),
            title="Select dds files to turn into bitmap tags")
        fps = [fp for fp in fps if fp.lower().endswith(".dds")]

    if not fps:
        return

    print("Creating bitmaps from dds files")
    for fp in sorted(fps):
        # make the tag window
        window = app.load_tags(filepaths='', def_id='bitm')
        if not window:
            return

        window = window[0]
        window.is_new_tag = True

        compile_bitmap_from_dds_files(window.tag, (Path(fp), ))
        window.update_title(list(Path(fp).parts)[-1])

        # reload the window to display the newly entered info
        window.reload()
        # prompt the user to save the tag somewhere
        app.save_tag_as()


def bitmap_from_multiple_dds(app, fps=()):
    load_dir = app.bitmap_load_dir
    data_dir = app.data_dir
    if is_path_empty(data_dir):
        data_dir = Path("")
    if is_path_empty(load_dir):
        load_dir = data_dir

    if not fps:
        fps = askopenfilenames(
            initialdir=load_dir, parent=app,
            filetypes=(("DDS image", "*.dds"), ("All", "*")),
            title="Select dds files to turn into a single bitmap tag")
        fps = [fp for fp in fps if fp.lower().endswith(".dds")]

    if not fps:
        return

    # make the tag window
    window = app.load_tags(filepaths='', def_id='bitm')
    if not window:
        return

    print("Creating bitmap from dds files")
    window = window[0]
    window.is_new_tag = True

    fps = sorted(fps)

    for fp in fps:
        pure_path = Path(fp)
        window.update_title(list(pure_path.parts)[-1])
        app.bitmap_load_dir = pure_path.parent
        break

    compile_bitmap_from_dds_files(window.tag, fps)

    # reload the window to display the newly entered info
    window.reload()
    # prompt the user to save the tag somewhere
    app.save_tag_as()


def bitmap_from_bitmap_source(app, e=None):
    load_dir = app.bitmap_load_dir
    if is_path_empty(load_dir):
        load_dir = app.last_data_load_dir

    fps = askopenfilenames(initialdir=load_dir, parent=app,
                           filetypes=(("bitmap", "*.bitmap"), ("All", "*")),
                           title="Select a bitmap tag to get the source tiff")

    if not fps:
        return

    app.bitmap_load_dir = os.path.dirname(fps[0])

    print('Creating bitmap from uncompressed source image of these bitmaps:')
    for fp in fps:
        print("  %s" % fp)

        width, height, pixels = extract_bitmap_tiff_data(fp)
        if not pixels:
            continue

        # make the tag window
        try:
            window = app.load_tags(filepaths='', def_id='bitm')
        except LookupError:
            print('    Could not make a new bitmap. Change the tag set.')
        if not window:
            continue

        window = window[0]
        window.is_new_tag = True

        add_bitmap_to_bitmap_tag(
            window.tag, width, height, 1, "texture_2d", "a8r8g8b8", 0, pixels)

        window.tag.rel_filepath = "untitled%s.bitmap" % app.untitled_num
        window.tag.filepath = app.tags_dir.joinpath(window.tag.rel_filepath)

        app.update_tag_window_title(window)

        # reload the window to display the newly entered info
        window.reload()
        # prompt the user to save the tag somewhere
        app.save_tag_as()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os
import sys
import tkinter as tk

from pathlib import Path
from supyr_struct.util import path_split, is_in_dir
from threading import Thread
from time import time
from traceback import format_exc

from binilla.widgets.binilla_widget import BinillaWidget
from binilla.windows.filedialog import askopenfilename, askdirectory

from reclaimer.halo_script.hsc import get_h1_scenario_script_object_type_strings

from mozzarilla import editor_constants as e_c


class DataExtractionWindow(tk.Toplevel, BinillaWidget):
    app_root = None
    handler = None
    print_interval = 5

    _extracting = False
    stop_extracting = False
    tag_data_extractors = ()

    listbox_index_to_def_id = ()
    tag_class_fcc_to_ext = ()
    tag_class_ext_to_fcc = ()

    def __init__(self, app_root, *args, **kwargs):
        self.handler = app_root.handler
        self.app_root = app_root
        kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)
        tk.Toplevel.__init__(self, app_root, *args, **kwargs)
        BinillaWidget.__init__(self, app_root, *args, **kwargs)

        self.tag_data_extractors = getattr(
            self.handler, "tag_data_extractors", None)
        self.tag_class_fcc_to_ext = {k: self.handler.id_ext_map[k].strip(".")
                                     for k in self.handler.id_ext_map}
        self.tag_class_ext_to_fcc = {self.tag_class_fcc_to_ext[k]: k
                                     for k in self.tag_class_fcc_to_ext}


        self.title("Tag Data Extractor")
        self.resizable(0, 0)
        self.update()
        try:
            self.iconbitmap(e_c.MOZZ_ICON_PATH)
        except Exception:
            print("Could not load window icon.")

        self.listbox_index_to_def_id = list(sorted(
            k for k in self.tag_data_extractors.keys()
            if k in self.tag_class_fcc_to_ext))

        # make the tkinter variables
        self.dir_path = tk.StringVar(self, self.handler.tagsdir)
        self.tag_path = tk.StringVar(self)
        self.overwrite = tk.BooleanVar(self)
        self.decode_adpcm = tk.BooleanVar(self)
        self.use_scenario_names_in_scripts = tk.BooleanVar(self)

        # make the frames
        self.options_frame = tk.LabelFrame(
            self, text="Extraction options")
        self.dir_extract_frame = tk.LabelFrame(
            self, text="Directory to extract from")
        self.dir_path_frame = tk.Frame(self.dir_extract_frame)
        self.def_ids_frame = tk.LabelFrame(
            self.dir_extract_frame, text="Select which tag types to extract from")
        self.tag_path_frame = tk.LabelFrame(
            self, text="Tag to extract from")

        self.def_ids_scrollbar = tk.Scrollbar(
            self.def_ids_frame, orient="vertical")
        self.def_ids_listbox = tk.Listbox(
            self.def_ids_frame, selectmode='multiple', highlightthickness=0,
            yscrollcommand=self.def_ids_scrollbar.set, exportselection=False)
        self.def_ids_scrollbar.config(command=self.def_ids_listbox.yview)

        for def_id in self.listbox_index_to_def_id:
            try:
                tag_ext = self.handler.id_ext_map[def_id].split('.')[-1]
            except KeyError:
                # not available with the current tag set
                continue
            self.def_ids_listbox.insert('end', tag_ext)
            self.def_ids_listbox.select_set('end')


        self.overwrite_cbtn = tk.Checkbutton(
            self.options_frame, variable=self.overwrite,
            text="Overwrite existing data files")
        self.decode_adpcm_cbtn = tk.Checkbutton(
            self.options_frame, variable=self.decode_adpcm,
            text="Decode Xbox ADPCM audio")
        self.use_scenario_names_in_scripts_cbtn = tk.Checkbutton(
            self.options_frame, variable=self.use_scenario_names_in_scripts,
            text="Use scenario names in scripts")

        self.dir_path_entry = tk.Entry(
            self.dir_path_frame, textvariable=self.dir_path)
        self.dir_browse_button = tk.Button(
            self.dir_path_frame, text="Browse", command=self.dir_browse)
        self.dir_extract_button = tk.Button(
            self.dir_path_frame, text="Extract", command=self.dir_extract)

        self.tag_path_entry = tk.Entry(
            self.tag_path_frame, textvariable=self.tag_path)
        self.tag_browse_button = tk.Button(
            self.tag_path_frame, text="Browse", command=self.tag_browse)
        self.tag_extract_button = tk.Button(
            self.tag_path_frame, text="Extract", command=self.tag_extract)

        self.cancel_extraction_button = tk.Button(
            self, text="Cancel extraction", command=self.cancel_extraction)


        self.overwrite_cbtn.pack(fill="x", anchor="nw", side="left")
        self.decode_adpcm_cbtn.pack(fill="x", anchor="nw")
        self.use_scenario_names_in_scripts_cbtn.pack(fill="x", anchor="nw", side="left")


        self.def_ids_listbox.pack(side='left', fill="both", expand=True)
        self.def_ids_scrollbar.pack(side='left', fill="y")


        self.dir_path_entry.pack(
            padx=(4, 0), pady=2, side='left', expand=True, fill='x')
        self.dir_browse_button.pack(padx=(0, 4), pady=2, side='left')
        self.dir_extract_button.pack(padx=(0, 4), pady=2, side='left')

        self.tag_path_entry.pack(
            padx=(4, 0), pady=2, side='left', expand=True, fill='x')
        self.tag_browse_button.pack(padx=(0, 4), pady=2, side='left')
        self.tag_extract_button.pack(padx=(0, 4), pady=2, side='left')

        self.dir_path_frame.pack(fill='x', padx=1)
        self.def_ids_frame.pack(fill='x', padx=1, expand=True)

        self.options_frame.pack(fill='x', padx=1)
        self.dir_extract_frame.pack(fill='x', padx=1)
        self.tag_path_frame.pack(fill='x', padx=1)
        self.cancel_extraction_button.pack(fill='both', padx=1, expand=True)

        self.transient(app_root)
        self.apply_style()
        self.update()
        w, h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry("%sx%s" % (w, h))
        self.minsize(width=w, height=h)

    def dir_browse(self):
        if self._extracting:
            return
        dirpath = askdirectory(
            initialdir=self.dir_path.get(),
            parent=self, title="Select the directory of tags to extract from")

        if not dirpath:
            return

        self.app_root.last_load_dir = dirpath
        if not is_in_dir(dirpath, self.handler.tagsdir):
            print('Directory "%s" is not located inside tags dir: "%s"'
                  % (dirpath, self.handler.tagsdir))
            return

        self.dir_path.set(dirpath)

    def tag_browse(self):
        if self._extracting:
            return
        filetypes = [('All', '*')]

        for def_id in sorted(self.tag_data_extractors.keys()):
            if def_id in self.tag_class_fcc_to_ext:
                filetypes.append(
                    (def_id, "." + self.tag_class_fcc_to_ext[def_id]))

        fp = askopenfilename(
            initialdir=str(self.app_root.last_load_dir), filetypes=filetypes,
            parent=self, title="Select a tag to extract from")

        if not fp:
            return

        fp = Path(fp)
        self.app_root.last_load_dir = fp.parent
        if not is_in_dir(fp, self.handler.tagsdir):
            print("Tag %s is not located in tags directory %s"
                  % (fp, self.handler.tagsdir))
            return

        self.app_root.last_load_dir = fp.parent
        self.tag_path.set(fp)

    def destroy(self):
        try:
            self.app_root.tool_windows.pop(self.window_name, None)
        except AttributeError:
            pass
        self.stop_extracting = True
        tk.Toplevel.destroy(self)

    def cancel_extraction(self):
        self.stop_extracting = True

    def get_tag(self, tag_path):
        def_id = self.handler.get_def_id(tag_path)
        try:
            tag = self.handler.get_tag(tag_path, def_id)
        except (KeyError, LookupError):
            tag = None

        try:
            if tag is None:
                return self.handler.build_tag(
                    filepath=self.handler.tagsdir.joinpath(tag_path))
        except Exception:
            pass
        return tag

    def dir_extract(self):
        if self._extracting:
            return
        try: self.extract_thread.join(1.0)
        except Exception: pass
        self.extract_thread = Thread(target=self._dir_extract, daemon=True)
        self.extract_thread.start()

    def tag_extract(self):
        if self._extracting:
            return
        try: self.extract_thread.join(1.0)
        except Exception: pass
        self.extract_thread = Thread(target=self._tag_extract, daemon=True)
        self.extract_thread.start()

    def _dir_extract(self):
        self._extracting = True
        self.stop_extracting = False
        try:
            self.do_dir_extract()
        except Exception:
            print(format_exc())
        self._extracting = False

    def _tag_extract(self):
        self._extracting = True
        self.stop_extracting = False
        try:
            self.do_tag_extract()
        except Exception:
            print(format_exc())
        self._extracting = False

    def do_dir_extract(self):
        tags_path = self.dir_path.get()
        data_path = self.handler.tagsdir.parent.joinpath("data")

        if not is_in_dir(tags_path, self.handler.tagsdir):
            print("Directory %s is not located inside tags dir: %s"
                  % (tags_path, self.handler.tagsdir))
            return

        settings = dict(out_dir=str(data_path), overwrite=self.overwrite.get(),
                        decode_adpcm=self.decode_adpcm.get(), engine="yelo")

        print("Beginning tag data extracton in:\t%s" % self.handler.tagsdir)

        s_time = time()
        c_time = s_time
        p_int = self.print_interval

        all_tag_paths = {self.listbox_index_to_def_id[int(i)]: [] for i in
                         self.def_ids_listbox.curselection()}

        print("Locating tags...")

        for root, directories, files in os.walk(tags_path):
            root = Path(root)
            try:
                root = root.relative_to(self.handler.tagsdir)
            except ValueError:
                continue

            for filename in files:
                filepath = root.joinpath(filename)

                if time() - c_time > p_int:
                    c_time = time()
                    print(' '*4, filepath, sep="")
                    self.app_root.update_idletasks()

                if self.stop_extracting:
                    print('Tag data extraction cancelled.\n')
                    return

                tag_paths = all_tag_paths.get(
                    self.tag_class_ext_to_fcc.get(
                        filepath.suffix[1:].lower()))

                if tag_paths is not None:
                    tag_paths.append(filepath)

        for def_id in sorted(all_tag_paths):
            extractor = self.tag_data_extractors[def_id]
            if self.stop_extracting:
                print('Tag data extraction cancelled.\n')
                return

            print("Extracting %s" % def_id)
            for filepath in all_tag_paths[def_id]:
                if self.stop_extracting:
                    print('Tag data extraction cancelled.\n')
                    return

                print(' '*4, filepath, sep="")
                self.extract(filepath, extractor, **settings)

        print("Extraction completed.\n")

    def do_tag_extract(self):
        tag_path = self.tag_path.get()

        def_id = self.tag_class_ext_to_fcc.get(Path(tag_path).suffix[1:].lower())
        if def_id is None or def_id not in self.tag_data_extractors:
            print("Cannot extract data from this kind of tag.")
            return
        else:
            if not is_in_dir(tag_path, self.handler.tagsdir):
                print("Tag %s is not located within tags directory: %s"
                      % (tag_path, self.handler.tagsdir))
                return

        print("Extracting %s" % tag_path)
        self.extract(os.path.relpath(tag_path, str(self.handler.tagsdir)),
                     self.tag_data_extractors[def_id],
                     out_dir=str(Path(path_split(self.handler.tagsdir, "tags")).joinpath("data")),
                     overwrite=self.overwrite.get(), engine="yelo",
                     decode_adpcm=self.decode_adpcm.get()
                     )

        print("Extraction completed.\n")

    def extract(self, tag_path, extractor, **settings):
        if extractor is None:
            print("No extractor defined for this kind of tag.")
            return

        tag = self.get_tag(tag_path)
        if tag is None:
            print((' '*8) + "Could not load tag.")
            return

        try:
            if (self.use_scenario_names_in_scripts.get() and
                tag.data[0].tag_class.enum_name == "scenario"):
                settings["hsc_node_strings_by_type"] = get_h1_scenario_script_object_type_strings(tag.data.tagdata)
        except Exception:
            print(format_exc())

        try:
            result = extractor(
                tag.data.tagdata, str(tag_path), byteswap_pcm_samples=True,
                **settings)
        except Exception:
            result = format_exc()

        if result:
            print(result)



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import tkinter as tk
import weakref

from array import array

from binilla.widgets.scroll_menu import ScrollMenu
from binilla.widgets.field_widgets import ContainerFrame
from binilla.widgets.bitmap_display_frame import BitmapDisplayFrame,\
     BitmapDisplayButton

from reclaimer.constants import CUBEMAP_PADDING, TYPE_NAME_MAP, FORMAT_NAME_MAP

try:
    import arbytmap
except ImportError:
    arbytmap = None

try:
    from reclaimer.bitmaps.p8_palette import HALO_P8_PALETTE, STUBBS_P8_PALETTE
except ImportError:
    HALO_P8_PALETTE = STUBBS_P8_PALETTE = None

SPRITE_RECTANGLE_TAG = "SPRITE_RECTANGLE"
SPRITE_CENTER_TAG = "SPRITE_CENTER"


class HaloBitmapDisplayBase:
    cubemap_padding = CUBEMAP_PADDING

    @property
    def engine(self):
        if getattr(self.master, "engine", None):
            return self.master.engine

        return getattr(getattr(self.master, "tag_window", None), "engine", None)

    def get_p8_palette(self, tag):
        if getattr(tag, "p8_palette", None):
            return tag.p8_palette
        elif not self.engine:
            return self.master.tag_window.tag.p8_palette
        elif "stubbs" in self.engine:
            return STUBBS_P8_PALETTE
        else:
            return HALO_P8_PALETTE

    def get_base_address(self, tag):
        if tag is None:
            return 0

        for b in tag.data.tagdata.bitmaps.STEPTREE:
            if b.pixels_offset:
                return b.pixels_offset
        return 0

    def is_xbox_bitmap(self, bitmap):
        try:
            return bitmap.base_address == 1073751810
        except AttributeError:
            return False

    def get_bitmap_pixels(self, bitmap_index, tag):
        bitmap = tag.data.tagdata.bitmaps.STEPTREE[bitmap_index]
        is_xbox = self.is_xbox_bitmap(bitmap)
        is_meta_tag = not hasattr(tag, "tags_dir")

        pixel_data = tag.data.tagdata.processed_pixel_data.data
        w, h, d = bitmap.width, bitmap.height, bitmap.depth
        fmt = FORMAT_NAME_MAP[bitmap.format.data]

        off = bitmap.pixels_offset
        if is_meta_tag:
            off -= self.get_base_address(tag)

        # xbox bitmaps are stored all mip level faces first, then
        # the next mip level, whereas pc is the other way. Xbox
        # bitmaps also have padding between each mipmap and bitmap.
        mipmap_count = bitmap.mipmaps + 1
        bitmap_count = 6 if bitmap.type.enum_name == "cubemap" else 1
        i_max = bitmap_count if is_xbox else mipmap_count
        j_max = mipmap_count if is_xbox else bitmap_count
        tex_block = []
        for i in range(i_max):
            if not is_xbox: mw, mh, md = arbytmap.get_mipmap_dimensions(w, h, d, i)

            for j in range(j_max):
                if is_xbox: mw, mh, md = arbytmap.get_mipmap_dimensions(w, h, d, j)

                if fmt == arbytmap.FORMAT_P8_BUMP:
                    tex_block.append(array('B', pixel_data[off: off + mw*mh]))
                    off += len(tex_block[-1])
                else:
                    off = arbytmap.bitmap_io.bitmap_bytes_to_array(
                        pixel_data, off, tex_block, fmt, mw, mh, md)

            # skip the xbox alignment padding to get to the next texture
            if is_xbox:
                size, mod = off, self.cubemap_padding
                off += (mod - (size % mod)) % mod

        return tex_block

    def get_textures(self, tag):
        if tag is None: return ()

        bitmaps = tag.data.tagdata.bitmaps.STEPTREE
        textures = []
        for i in range(len(bitmaps)):
            b = bitmaps[i]
            typ = TYPE_NAME_MAP[0]
            fmt = FORMAT_NAME_MAP[0]
            if b.type.data in range(len(TYPE_NAME_MAP)):
                typ = TYPE_NAME_MAP[b.type.data]

            if b.format.data in range(len(FORMAT_NAME_MAP)):
                fmt = FORMAT_NAME_MAP[b.format.data]

            tex_info = dict(
                width=b.width, height=b.height, depth=b.depth,
                format=fmt, texture_type=typ,
                sub_bitmap_count=6 if typ == "CUBE" else 1,
                swizzled=b.flags.swizzled, mipmap_count=b.mipmaps,
                reswizzler="MORTON", deswizzler="MORTON")

            mipmap_count = b.mipmaps + 1
            if fmt == arbytmap.FORMAT_P8_BUMP:
                p8_palette = self.get_p8_palette(tag)
                tex_info.update(
                    palette=[p8_palette.p8_palette_32bit_packed]*mipmap_count,
                    palette_packed=True, indexing_size=8)

            tex_block = self.get_bitmap_pixels(i, tag)
            if self.is_xbox_bitmap(b) and typ == "CUBE":
                template = tuple(tex_block)
                i = 0
                for f in (0, 2, 1, 3, 4, 5):
                    for m in range(0, (b.mipmaps + 1)*6, 6):
                        tex_block[m + f] = template[i]
                        i += 1

            textures.append((tex_block, tex_info))

        return textures


class HaloBitmapDisplayFrame(BitmapDisplayFrame):
    # these mappings have the 2nd and 3rd faces swapped on pc for some reason
    cubemap_cross_mapping = (
        (-1,  1),
        ( 2,  4,  0,  5),
        (-1,  3),
        )

    def __init__(self, master, bitmap_tag=None, *args, **kwargs):
        self.bitmap_tag = bitmap_tag
        textures = kwargs.get('textures', ())
        BitmapDisplayFrame.__init__(self, master, *args, **kwargs)
        self.sequence_index = tk.IntVar(self)
        self.sprite_index   = tk.IntVar(self)

        labels = []
        labels.append(tk.Label(self.controls_frame0, text="Sequence index"))
        labels.append(tk.Label(self.controls_frame1, text="Sprite index"))
        for lbl in labels:
            lbl.config(width=15, anchor='w')

        self.sequence_menu = ScrollMenu(self.controls_frame0, menu_width=7,
                                        variable=self.sequence_index)
        self.sprite_menu = ScrollMenu(self.controls_frame1, menu_width=7,
                                      variable=self.sprite_index)

        padx = self.horizontal_padx
        pady = self.horizontal_pady
        for lbl in labels:
            lbl.pack(side='left', padx=(15, 0), pady=pady)
        self.sequence_menu.pack(side='left', padx=padx, pady=pady)
        self.sprite_menu.pack(side='left', padx=padx, pady=pady)
        self.write_trace(self.sequence_index, self.sequence_changed)
        self.write_trace(self.sprite_index,   self.sprite_changed)

        self.change_textures(textures)
        self.apply_style()

    def sequence_changed(self, *args):
        tag = self.bitmap_tag
        if tag is None:
            self.sprite_menu.set_options(())
            self.sequence_menu.set_options(("None", ))
            self.sprite_menu.sel_index = -1
            self.sequence_menu.sel_index = 0
            return

        sequence_i = self.sequence_index.get() - 1
        options = ()
        sequences = tag.data.tagdata.sequences.STEPTREE
        if sequence_i in range(len(sequences)):
            sequence = sequences[sequence_i]
            options = range(len(sequence.sprites.STEPTREE))
            if not options:
                options = range(sequence.bitmap_count)

        self.sprite_menu.set_options(options)
        self.sprite_menu.sel_index = (self.sprite_menu.max_index >= 0) - 1

    def sprite_changed(self, *args):
        self.image_canvas.delete(SPRITE_RECTANGLE_TAG)
        self.image_canvas.delete(SPRITE_CENTER_TAG)
        tag = self.bitmap_tag
        if tag is None:
            self.sprite_menu.set_options(())
            self.sequence_menu.set_options(("None", ))
            self.sprite_menu.sel_index = -1
            self.sequence_menu.sel_index = 0
            return

        data = tag.data.tagdata
        sequences = data.sequences.STEPTREE
        bitmaps   = data.bitmaps.STEPTREE

        sequence_i = self.sequence_index.get() - 1
        if sequence_i not in range(len(sequences)):
            return

        sequence = sequences[sequence_i]
        sprite_i = self.sprite_index.get()
        sprites = sequence.sprites.STEPTREE
        bitmap_index = sequence.first_bitmap_index + sprite_i
        x0, y0, x1, y1 = 0, 0, 1, 1
        rx, ry = 0.5, 0.5
        if sprite_i in range(len(sprites)):
            sprite = sprites[sprite_i]
            if sprite.bitmap_index not in range(len(bitmaps)):
                return

            bitmap_index = sprite.bitmap_index
            x0, y0, x1, y1 = sprite.left_side,  sprite.top_side,\
                             sprite.right_side, sprite.bottom_side
            rx = x0 + sprite.registration_point_x
            ry = y0 + sprite.registration_point_y
        elif bitmap_index not in range(len(bitmaps)):
            return

        if bitmap_index != self.bitmap_index.get():
            self.bitmap_menu.sel_index = bitmap_index

        bitmap = bitmaps[bitmap_index]
        mip = self.mipmap_index.get()
        w, h = max(bitmap.width>>mip, 1), max(bitmap.height>>mip, 1)
        x0, y0 = int(round(x0 * w)), int(round(y0 * h))
        x1, y1 = int(round(x1 * w)), int(round(y1 * h))
        rx, ry = int(round(rx * w)), int(round(ry * h))
        rx0, rx1 = rx - 1, rx + 1
        ry0, ry1 = ry - 1, ry + 1
        if x1 < x0: x0, x1 = x1, x0
        if y1 < y0: y0, y1 = y1, y0
        if ry1 < ry0: ry0, ry1 = ry1, ry0
        x0 -= 1; y0 -= 1
        rx0 -= 1; ry0 -= 1

        self.image_canvas.create_rectangle(
            (x0, y0, x1, y1), fill=None, dash=(2, 1),
            tags=SPRITE_RECTANGLE_TAG,
            outline=self.bitmap_canvas_outline_color)

        self.image_canvas.create_rectangle(
            (rx0, ry0, rx1, ry1), fill=None,
            tags=SPRITE_CENTER_TAG,
            outline=self.bitmap_canvas_bg_color)

    def change_textures(self, textures):
        BitmapDisplayFrame.change_textures(self, textures)
        tag = self.bitmap_tag
        if tag is None: return

        data = tag.data.tagdata
        options = {i+1: str(i) for i in range(len(data.sequences.STEPTREE))}
        options[0] = "None"

        if not (hasattr(self, "sprite_menu") and
                hasattr(self, "sequence_menu")):
            return
        self.sequence_menu.set_options(options)
        self.sequence_menu.sel_index = (self.sequence_menu.max_index >= 0) - 1


class HaloBitmapDisplayButton(HaloBitmapDisplayBase, BitmapDisplayButton):
    tags_dir = ""
    display_frame_class = HaloBitmapDisplayFrame

    def __init__(self, *args, **kwargs):
        self.tags_dir = kwargs.pop("tags_dir", self.tags_dir)
        BitmapDisplayButton.__init__(self, *args, **kwargs)

    def show_window(self, e=None, parent=None):
        w = tk.Toplevel()
        tag = self.bitmap_tag
        self.display_frame = weakref.ref(self.display_frame_class(w, tag))
        self.display_frame().change_textures(self.get_textures(self.bitmap_tag))
        self.display_frame().pack(expand=True, fill="both")

        try:
            tag_name = "untitled"
            tag_name = tag.filepath.lower()
            tag_name = tag_name.split(self.tags_dir.lower(), 1)[-1]
        except Exception:
            pass
        w.title("Preview: %s" % tag_name)
        w.transient(parent)
        w.focus_force()
        return w


class HaloBitmapTagFrame(ContainerFrame):
    bitmap_display_button_class = HaloBitmapDisplayButton

    def bitmap_preview(self, e=None):
        f = self.preview_btn.display_frame
        if f is not None and f() is not None:
            return

        try:
            tag = self.tag_window.tag
        except AttributeError:
            return
        self.preview_btn.change_bitmap(tag)
        self.preview_btn.show_window(None, self.tag_window.app_root)

    def pose_fields(self):
        try:
            tags_dir = self.tag_window.tag.tags_dir
        except AttributeError:
            tags_dir = ""
        self.preview_btn = self.bitmap_display_button_class(
            self, width=15, text="Preview",
            tags_dir=tags_dir, command=self.bitmap_preview,)

        self.preview_btn.pack(anchor='center', pady=10)
        ContainerFrame.pose_fields(self)



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from binilla.widgets.bitmap_display_frame import BitmapDisplayFrame
from mozzarilla.widgets.field_widgets.halo_1_bitmap_display import \
     HaloBitmapDisplayFrame, HaloBitmapDisplayButton, HaloBitmapTagFrame
from reclaimer.constants import TYPE_NAME_MAP, FORMAT_NAME_MAP
from reclaimer.h3.util import get_virtual_dimension, get_h3_pixel_bytes_size

try:
    import arbytmap
except ImportError:
    arbytmap = None


class Halo3BitmapDisplayFrame(HaloBitmapDisplayFrame):
    cubemap_cross_mapping = BitmapDisplayFrame.cubemap_cross_mapping

    def _display_2d_bitmap(self, force=False, bitmap_mapping=None):
        image_ct = len(self.get_images())
        if bitmap_mapping is None and self.active_image_handler.tex_type == "2D":
            bitmap_mapping = []
            column_ct = image_ct // 4
            for i in range(4 * (column_ct > 1)):
                bitmap_mapping.append(list(i * column_ct + j for
                                           j in range(column_ct)))

            if column_ct * 4 < image_ct:
                bitmap_mapping.append(list(range(column_ct * 4, image_ct)))

        HaloBitmapDisplayFrame._display_2d_bitmap(self, force, bitmap_mapping)


class Halo3BitmapDisplayButton(HaloBitmapDisplayButton):
    display_frame_class = Halo3BitmapDisplayFrame

    def get_base_address(self, tag):
        return 0

    def get_bitmap_pixels(self, bitmap_index, tag):
        bitmap = tag.data.tagdata.bitmaps.STEPTREE[bitmap_index]
        is_meta_tag = not hasattr(tag, "tags_dir")

        off = bitmap.pixels_offset
        pixel_data = tag.data.tagdata.processed_pixel_data.data

        w, h, d = bitmap.width, bitmap.height, bitmap.depth
        tiled = bitmap.format_flags.tiled
        fmt = FORMAT_NAME_MAP[bitmap.format.data]

        # xbox bitmaps are stored all mip level faces first, then
        # the next mip level, whereas pc is the other way. Xbox
        # bitmaps also have padding between each mipmap and bitmap.
        bitmap_count = 1
        mipmap_count = bitmap.mipmaps + 1
        if bitmap.type.enum_name == "cubemap":
            bitmap_count = 6
        elif bitmap.type.enum_name == "multipage_2d":
            bitmap_count = d
            d = 1

        tex_block = []
        if fmt == arbytmap.FORMAT_P8_BUMP:
            fmt = arbytmap.FORMAT_A8

        for i in range(mipmap_count):
            pixel_data_size = get_h3_pixel_bytes_size(fmt, w, h, d, i, tiled)
            for j in range(bitmap_count):
                off = arbytmap.bitmap_io.bitmap_bytes_to_array(
                    pixel_data, off, tex_block, fmt,
                    1, 1, 1, pixel_data_size)

        return tex_block

    def get_textures(self, tag):
        textures = HaloBitmapDisplayButton.get_textures(self, tag)

        bitmaps = tag.data.tagdata.bitmaps.STEPTREE
        for i in range(len(bitmaps)):
            bitmap = bitmaps[i]
            tex_info = textures[i][1]

            if bitmap.type.enum_name == "multipage_2d":
                tex_info.update(depth=1, texture_type=TYPE_NAME_MAP[0],
                                sub_bitmap_count=tex_info["depth"])

            # update the texture info
            tex_info.update(
                packed_width_calc=get_virtual_dimension,
                packed_height_calc=get_virtual_dimension,
                target_packed_width_calc=get_virtual_dimension,
                target_packed_height_calc=get_virtual_dimension,
                big_endian=True, target_big_endian=True,
                tiled=bitmap.format_flags.tiled,
                tile_mode=False, tile_method="DXGI")

        return textures


class Halo3BitmapTagFrame(HaloBitmapTagFrame):
    bitmap_display_button_class = Halo3BitmapDisplayButton



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from traceback import format_exc

from binilla.widgets.field_widget_picker import WidgetPicker
from binilla.widgets.binilla_widget import BinillaWidget

from mozzarilla import editor_constants as e_c
from mozzarilla.widgets.field_widgets import DependencyFrame, EntryFrame
from reclaimer.field_types import StrTagRef, TagRef, ZoneAsset, StringID, \
     StrUtf16, FlStrUTF16Data, StrLatin1Enum, FlFloat, FlStrUTF16,\
     FlUInt16, FlSInt16, FlUEnum16, FlSEnum16, FlBool16,\
     FlUInt32, FlSInt32, FlUEnum32, FlSEnum32, FlBool32,\
     TagIndex, RawdataRef, Reflexive, RawReflexive
from supyr_struct.field_types import Struct, Array, Float, StrUtf16,\
     UInt16, SInt16, UEnum16, SEnum16, Bool16,\
     UInt32, SInt32, UEnum32, SEnum32, Bool32


BinillaWidget.title_width = e_c.TITLE_WIDTH
BinillaWidget.def_string_entry_width = e_c.DEF_STRING_ENTRY_WIDTH
BinillaWidget.max_string_entry_width = e_c.MAX_STRING_ENTRY_WIDTH

__all__ = ("WidgetPicker", "MozzarillaWidgetPicker", "def_halo_widget_picker")

class MozzarillaWidgetPicker(WidgetPicker):
    pass

def_halo_widget_picker = dhwp = MozzarillaWidgetPicker()

dhwp.add_widget(StrTagRef, EntryFrame)
dhwp.add_widget(TagRef, DependencyFrame)

dhwp.copy_widget(FlStrUTF16Data, StrUtf16)
dhwp.copy_widget(FlStrUTF16, StrUtf16)

dhwp.copy_widget(FlUInt16, UInt16)
dhwp.copy_widget(FlUInt32, UInt32)
dhwp.copy_widget(FlUEnum16, UEnum16)
dhwp.copy_widget(FlUEnum32, UEnum32)
dhwp.copy_widget(FlBool16, Bool16)
dhwp.copy_widget(FlBool32, Bool32)
dhwp.copy_widget(FlSInt16, SInt16)
dhwp.copy_widget(FlSInt32, SInt32)
dhwp.copy_widget(FlSEnum16, SEnum16)
dhwp.copy_widget(FlSEnum32, SEnum32)

dhwp.copy_widget(FlFloat, Float)

dhwp.copy_widget(ZoneAsset, Struct)
dhwp.copy_widget(StringID, Struct)
dhwp.copy_widget(TagIndex, Array)
dhwp.copy_widget(RawdataRef, Struct)
dhwp.copy_widget(Reflexive, Struct)
dhwp.copy_widget(RawReflexive, Struct)
dhwp.copy_widget(StrLatin1Enum, SEnum32)


#############################
#           Halo 2
#############################
try:
    from reclaimer.h2.field_types import H2TagRef, H2RawdataRef, H2Reflexive
    dhwp.copy_widget(H2TagRef, TagRef)
    dhwp.copy_widget(H2RawdataRef, RawdataRef)
    dhwp.copy_widget(H2Reflexive, Reflexive)
except Exception:
    print(format_exc())


#############################
#           Halo 3
#############################
try:
    from reclaimer.h3.field_types import H3TagRef, H3RawdataRef, H3Reflexive
    dhwp.copy_widget(H3TagRef, TagRef)
    dhwp.copy_widget(H3RawdataRef, RawdataRef)
    dhwp.copy_widget(H3Reflexive, Reflexive)
except Exception:
    print(format_exc())



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import reclaimer
import tkinter as tk

from reclaimer.strings.strings_decompilation import parse_hmt_message

from binilla.widgets.field_widgets.computed_text_frame import ComputedTextFrame


class HaloScriptTextFrame(ComputedTextFrame):
    syntax  = None
    strings = None

    def get_text(self):
        if self.parent is None:
            return ""

        if None in (self.strings, self.syntax):
            tag_data = self.parent.parent.parent.parent
            self.syntax  = reclaimer.halo_script.hsc.get_hsc_data_block(
                tag_data.script_syntax_data.data)
            self.strings = tag_data.script_string_data.data.decode("latin-1")

        if None in (self.strings, self.syntax):
            return

        typ = "script"
        if "global" in self.f_widget_parent.node.NAME:
            typ = "global"

        tag_data = self.parent.parent.parent.parent
        script_strings_by_type = ()
        try:
            if self.tag_window.use_scenario_names_for_script_names:
                script_strings_by_type = reclaimer.halo_script.hsc.\
                                         get_h1_scenario_script_object_type_strings(tag_data)
        except Exception:
            pass

        new_text = reclaimer.halo_script.hsc.hsc_bytecode_to_string(
                self.syntax, self.strings, self.f_widget_parent.attr_index,
                tag_data.scripts.STEPTREE, tag_data.globals.STEPTREE, typ,
                hsc_node_strings_by_type=script_strings_by_type)
        return new_text


class HaloHudMessageTextFrame(ComputedTextFrame):
    def get_text(self):
        if self.parent is None:
            return ""

        tag_data = self.parent.parent.parent.parent
        message_index = self.parent.parent.index(self.parent)
        return parse_hmt_message(tag_data, message_index)[0]



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = ("ObjectConverter", "ModelAnimationsConverter",
           "ModelConverter", "GbxmodelConverter", "CollisionConverter",
           "ChicagoShaderConverter", "SbspConverter")


from mozzarilla.windows.tag_converters.model_converter import ModelConverter
from mozzarilla.windows.tag_converters.model_animations_converter import ModelAnimationsConverter
from mozzarilla.windows.tag_converters.object_converter import ObjectConverter
from mozzarilla.windows.tag_converters.gbxmodel_converter import GbxmodelConverter
from mozzarilla.windows.tag_converters.chicago_shader_converter import ChicagoShaderConverter
from mozzarilla.windows.tag_converters.collision_converter import CollisionConverter
from mozzarilla.windows.tag_converters.sbsp_converter import SbspConverter



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = (
    "field_widgets", "field_widget_picker", "directory_frame"
    )

from mozzarilla.widgets import field_widgets, field_widget_picker,\
     directory_frame



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from binilla.defs.config_def import method_enums, modifier_enums, depths,\
     hotkey_enums, array_counts, app_window, tag_windows, tag_printing,\
     filepath, open_tags, recent_tags, directory_paths, theme_name,\
     tag_window_hotkeys, padding, widths_and_heights, tag_backup
from binilla.defs.style_def import appearance, color, font
from binilla.constants import GUI_NAME, NAME, TOOLTIP, VALUE,\
     VISIBILITY_METADATA, VISIBILITY_HIDDEN
from binilla.widgets.field_widgets.array_frame import DynamicArrayFrame
from binilla.defs import config_tooltips as ttip

from supyr_struct.defs.tag_def import TagDef
from supyr_struct.field_types import *

from mozzarilla.editor_constants import mozz_color_names, mozz_font_names


__all__ = (
    "get", "config_def",
    )

mozz_flag_tooltips = (
    "Whether to show the hierarchy window in the main window.",
    "Whether to show the console output in the main window.",
    ("Whether to recalculate certain hidden values when saving.\n" +
     "For all intents and purposes, this should stay on unless\n" +
     "you are doing some form of experimenting or debugging."),
    "When reading scripts, extract names for encounters,\n"
    "command lists, scripts, cutscene titles/camera points/flags,\n"
    "trigger volumes, recorded animations, ai conversations,\n"
    "object names, device groups, and player starting profiles\n"
    "from the scenarios reflexives, rather than script strings.",
    "Whether or not to show the full tags directory in the tag window title."
    )

new_method_enums = (
    # start at 1024 to make sure to provide space for any future binilla methods
    {GUI_NAME:"", NAME:"mozz_divider1", VALUE:1024},
    {GUI_NAME:"MOZZARILLA METHODS", NAME:"mozz_divider2"},
    {GUI_NAME:"choose tags directory", NAME:"set_tags_dir"},
    {GUI_NAME:"switch tags directory", NAME:"switch_tags_dir"},
    {GUI_NAME:"add tags directory",    NAME:"add_tags_dir"},
    {GUI_NAME:"remove tags directory", NAME:"remove_tags_dir"},
    # space for (64 - 6) more enums here

    {GUI_NAME:"", NAME:"mozz_divider3", VALUE:1024 + 64},
    {GUI_NAME:"open dependency scanner", NAME:"show_dependency_viewer"},
    {GUI_NAME:"open tag scanner", NAME:"show_tag_scanner"},
    {GUI_NAME:"open search and replace", NAME:"show_search_and_replace"},
    {GUI_NAME:"make bitmap(s) from dds", NAME:"bitmap_from_dds"},
    {GUI_NAME:"make bitmap from bitmap source", NAME:"bitmap_from_bitmap_source"},
    {GUI_NAME:"launch pool", NAME:"create_hek_pool_window"},
    {GUI_NAME:"open tag data extractor", NAME:"show_data_extraction_window"},
    {GUI_NAME:"make physics from jms", NAME:"physics_from_jms"},
    {GUI_NAME:"make gbxmodel from jms", NAME:"model_from_jms"},
    {GUI_NAME:"make hud_message_text from hmt", NAME:"hud_message_text_from_hmt"},
    {GUI_NAME:"make bitmap from dds", NAME:"bitmap_from_multiple_dds"},
    {GUI_NAME:"make strings from txt", NAME:"strings_from_txt"},
    {GUI_NAME:"open bitmap source extractor", NAME:"show_bitmap_source_extractor"},
    {GUI_NAME:"open model animations compiler", NAME:"show_animations_compiler_window"},
    {GUI_NAME:"open model animations compression", NAME:"show_animations_compression_window"},
    {GUI_NAME:"open sound compiler", NAME:"show_sound_compiler_window"},
    # space for (64 - 17) more enums here

    {GUI_NAME:"", NAME:"mozz_divider4", VALUE:1024 + 64*2},
    {GUI_NAME:"open model converter", NAME:"show_model_converter"},
    {GUI_NAME:"open gbxmodel converter", NAME:"show_gbxmodel_converter"},
    {GUI_NAME:"open collision converter", NAME:"show_collision_converter"},
    {GUI_NAME:"open sbsp converter", NAME:"show_sbsp_converter"},
    {GUI_NAME:"open chicago shader converter", NAME:"show_chicago_shader_converter"},
    {GUI_NAME:"open animations converter", NAME:"show_animations_converter"},
    {GUI_NAME:"open object converter", NAME:"show_object_converter"},
    )

method_enums += new_method_enums

mozz_colors = Array("colors",
    SUB_STRUCT=color, SIZE="array_counts.color_count",
    MAX=len(mozz_color_names), MIN=len(mozz_color_names),
    NAME_MAP=mozz_color_names,
    GUI_NAME="Colors"
    )

mozz_fonts = Array("fonts",
    SUB_STRUCT=font, SIZE="array_counts.font_count",
    MAX=len(mozz_font_names), MIN=len(mozz_font_names),
    NAME_MAP=mozz_font_names,
    GUI_NAME="Fonts"
    )

mozz_hotkey = Struct("hotkey",
    BitStruct("combo",
        UBitEnum("modifier", GUI_NAME="", *modifier_enums, SIZE=4),
        UBitEnum("key", GUI_NAME="and", *hotkey_enums, SIZE=28),
        SIZE=4, ORIENT='h', TOOLTIP=ttip.hotkey_combo
        ),
    UEnum32("method", *method_enums, TOOLTIP=ttip.hotkey_method)
    )

mozz_hotkeys = Array(
    "hotkeys", SUB_STRUCT=mozz_hotkey, DYN_NAME_PATH='.method.enum_name',
    SIZE="array_counts.hotkey_count", WIDGET=DynamicArrayFrame,
    GUI_NAME="Main window hotkeys"
    )

open_mozz_tag = Container("open_tag",
    Struct("header",
        UInt16("width"),
        UInt16("height"),
        SInt16("offset_x"),
        SInt16("offset_y"),
        Bool32("flags",
            "minimized",
            ),

        # UPDATE THIS PADDING WHEN ADDING STUFF ABOVE IT
        Pad(48 - 2*4 - 4*1),

        UInt16("def_id_len"),
        UInt16("path_len"),
        UInt16("tags_dir_index"),
        UInt16("handler_index"),
        SIZE=64
        ),

    StrUtf8("def_id", SIZE=".header.def_id_len"),
    StrUtf8("path", SIZE=".header.path_len"),
    )

mozzarilla = Container("mozzarilla",
    Bool16("flags",
        {NAME: "show_hierarchy_window", TOOLTIP: mozz_flag_tooltips[0]},
        {NAME: "show_console_window", TOOLTIP: mozz_flag_tooltips[1]},
        {NAME: "calc_internal_data", TOOLTIP: mozz_flag_tooltips[2]},
        {NAME: "use_scenario_names_in_scripts", TOOLTIP: mozz_flag_tooltips[3]},
        {NAME: "show_full_tags_directory", TOOLTIP: mozz_flag_tooltips[4]},
        DEFAULT=sum([1<<i for i in (0, 1, 2)])
        ),
    UEnum16("selected_handler",
        "halo_1",
        "halo_1_os_v3",
        "halo_1_os_v4",
        "halo_1_misc",
        "stubbs",
        "halo_3",
        EDITABLE=False, VISIBLE=VISIBILITY_METADATA
        ),
    UInt16("last_tags_dir",  VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt16("sash_position",  VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt32("last_tool_path", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    Pad(64 - 2*4 - 4*1),

    UInt16("tags_dirs_count",     VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt16("load_dirs_count",     VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt16("open_mozz_tag_count", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    Pad(64 - 2*3),

    Array("tags_dirs", SUB_STRUCT=filepath, SIZE="mozzarilla.tags_dirs_count",
        MIN=1, VISIBLE=VISIBILITY_METADATA),
    Array("load_dirs", SUB_STRUCT=filepath, SIZE="mozzarilla.load_dirs_count",
        NAME_MAP=("last_data_load_dir", "jms_load_dir", "bitmap_load_dir"),
        MIN=3, VISIBLE=VISIBILITY_METADATA
        ),
    Array("open_mozz_tags",
        SUB_STRUCT=open_mozz_tag, SIZE="mozzarilla.open_mozz_tag_count", VISIBLE=VISIBILITY_HIDDEN
        ),
    COMMENT="\nThese are settings specific to Mozzarilla.",
    GUI_NAME="Mozzarilla"
    )

mozz_version_info = Struct("version_info",
    UEnum32("id", ('Mozz', 'zzoM'), VISIBLE=VISIBILITY_METADATA, DEFAULT='zzoM'),
    UInt32("version", DEFAULT=3, VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    Timestamp32("date_created", EDITABLE=False),
    Timestamp32("date_modified", EDITABLE=False),
    SIZE=16, VISIBLE=VISIBILITY_HIDDEN
    )

mozz_appearance = Container("appearance",
    theme_name,
    widths_and_heights,
    padding,
    depths,
    mozz_colors,
    mozz_fonts,
    GUI_NAME="Appearance", COMMENT=(
        "\nThese settings control how everything looks. Colors, fonts, etc."
        "\nThese settings are what get saved to/loaded from style files.")
    )

mozz_all_hotkeys = Container("all_hotkeys",
    mozz_hotkeys,
    tag_window_hotkeys,
    GUI_NAME="Hotkeys", COMMENT=(
        "\nThese hotkeys control what operations to bind to keystroke"
        "\ncombinations for the main window and the tag windows.")
    )

config_def = TagDef("mozzarilla_config",
    mozz_version_info,  # not visible
    array_counts,  # not visible
    app_window,
    tag_windows,
    tag_printing,
    tag_backup,
    open_tags,  # not visible
    recent_tags,  # not visible
    directory_paths,  # not visible
    mozz_appearance,
    mozz_all_hotkeys,

    mozzarilla,
    ENDIAN='<', ext=".cfg",
    )

mozz_config_version_def = TagDef(mozz_version_info)

def get(): return config_def



#!/usr/bin/env python3
#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

# Legacy run module, used by older MEKs

from .__main__ import main
if main():
    # Input was how the terminal window was kept open on Windows to show the
    # error.
    input()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import ctypes
import os
import sys
import tkinter as tk

from pathlib import Path
from time import time
from threading import Thread
from traceback import format_exc

from binilla.util import ProcController, open_in_default_program
from binilla.widgets.binilla_widget import BinillaWidget
from binilla.windows.filedialog import askdirectory, asksaveasfilename

from supyr_struct.util import path_normalize, is_in_dir

from mozzarilla import editor_constants as e_c


platform = sys.platform.lower()
if "linux" in platform:
    platform = "linux"


SetFileAttributesW = None
if platform == "win32":
    SetFileAttributesW = ctypes.windll.kernel32.SetFileAttributesW


class TagScannerWindow(tk.Toplevel, BinillaWidget):
    app_root = None
    handler = None

    _scanning = False
    stop_scanning = False
    print_interval = 5

    listbox_index_to_def_id = ()

    def __init__(self, app_root, *args, **kwargs):
        self.handler = handler = app_root.handler
        self.app_root = app_root
        kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)
        tk.Toplevel.__init__(self, app_root, *args, **kwargs)
        BinillaWidget.__init__(self, app_root, *args, **kwargs)

        self.title("[%s] Tags directory error locator" %
                   app_root.handler_names[app_root._curr_handler_index])
        self.minsize(width=400, height=300)
        self.update()
        try:
            self.iconbitmap(e_c.MOZZ_ICON_PATH)
        except Exception:
            print("Could not load window icon.")

        ext_id_map = handler.ext_id_map
        self.listbox_index_to_def_id = [
            ext_id_map[ext] for ext in sorted(ext_id_map.keys())]

        # make the tkinter variables
        self.open_logfile = tk.BooleanVar(self, True)
        self.directory_path = tk.StringVar(self)
        self.logfile_path = tk.StringVar(self)

        # make the frames
        self.directory_frame = tk.LabelFrame(self, text="Directory to scan")
        self.logfile_frame   = tk.LabelFrame(self, text="Output log filepath")
        self.def_ids_frame = tk.LabelFrame(
            self, text="Select which tag types to scan")
        self.logfile_dir_frame = tk.Frame(self.logfile_frame)
        self.button_frame = tk.Frame(self.def_ids_frame)

        self.scan_button = tk.Button(
            self.button_frame, text='Scan directory',
            width=20, command=self.scan_directory)
        self.cancel_button = tk.Button(
            self.button_frame, text='Cancel scan',
            width=20, command=self.cancel_scan)
        self.select_all_button = tk.Button(
            self.button_frame, text='Select all',
            width=20, command=self.select_all)
        self.deselect_all_button = tk.Button(
            self.button_frame, text='Deselect all',
            width=20, command=self.deselect_all)

        self.directory_entry = tk.Entry(
            self.directory_frame, textvariable=self.directory_path)
        self.dir_browse_button = tk.Button(
            self.directory_frame, text="Browse", command=self.dir_browse)

        self.logfile_entry = tk.Entry(
            self.logfile_dir_frame, textvariable=self.logfile_path,)
        self.log_browse_button = tk.Button(
            self.logfile_dir_frame, text="Browse", command=self.log_browse)
        self.open_logfile_cbtn = tk.Checkbutton(
            self.logfile_frame, text="Open log when done scanning",
            variable=self.open_logfile)

        self.def_ids_scrollbar = tk.Scrollbar(
            self.def_ids_frame, orient="vertical")
        self.def_ids_listbox = tk.Listbox(
            self.def_ids_frame, selectmode='multiple', highlightthickness=0,
            yscrollcommand=self.def_ids_scrollbar.set, exportselection=False)
        self.def_ids_scrollbar.config(command=self.def_ids_listbox.yview)

        for def_id in self.listbox_index_to_def_id:
            tag_ext = handler.id_ext_map[def_id].split('.')[-1]
            self.def_ids_listbox.insert('end', tag_ext)

            # these tag types are massive, so by
            # default dont set them to be scanned
            if def_id in ("sbsp", "scnr"):
                continue
            self.def_ids_listbox.select_set('end')

        for w in (self.directory_entry, self.logfile_entry):
            w.pack(padx=(4, 0), pady=2, side='left', expand=True, fill='x')

        for w in (self.dir_browse_button, self.log_browse_button):
            w.pack(padx=(0, 4), pady=2, side='left')

        for w in (self.scan_button, self.cancel_button):
            w.pack(padx=4, pady=2)

        for w in (self.deselect_all_button, self.select_all_button):
            w.pack(padx=4, pady=2, side='bottom')

        self.def_ids_listbox.pack(side='left', fill="both", expand=True)
        self.def_ids_scrollbar.pack(side='left', fill="y")
        self.button_frame.pack(side='left', fill="y")

        self.directory_frame.pack(fill='x', padx=1)
        self.logfile_frame.pack(fill='x', padx=1)
        self.logfile_dir_frame.pack(fill='x')
        self.open_logfile_cbtn.pack(fill='x', side=tk.LEFT)
        self.def_ids_frame.pack(fill='both', padx=1, expand=True)

        self.transient(app_root)

        self.directory_path.set(handler.tagsdir)
        self.logfile_path.set(Path(handler.tagsdir, "tag_scanner.log"))
        self.apply_style()
        self.update()
        w, h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry("%sx%s" % (w, h))
        self.minsize(width=w, height=h)

    def deselect_all(self):
        if self._scanning:
            return
        self.def_ids_listbox.select_clear(0, 'end')

    def select_all(self):
        if self._scanning:
            return
        for i in range(len(self.listbox_index_to_def_id)):
            self.def_ids_listbox.select_set(i)

    def get_tag(self, filepath):
        handler = self.handler
        def_id = handler.get_def_id(filepath)

        try:
            tag = handler.get_tag(filepath, def_id)
        except (KeyError, LookupError):
            tag = None
        try:
            if tag is None:
                return handler.build_tag(
                    filepath=self.handler.tagsdir.joinpath(filepath)
                    )
        except Exception:
            pass
        return tag

    def dir_browse(self):
        if self._scanning:
            return
        dirpath = askdirectory(initialdir=self.directory_path.get(),
                               parent=self, title="Select directory to scan")

        if not dirpath:
            return

        self.app_root.last_load_dir = dirpath
        if not is_in_dir(dirpath, self.handler.tagsdir):
            print("Specified directory is not located within the tags directory")
            return

        self.directory_path.set(dirpath)

    def log_browse(self):
        if self._scanning:
            return
        filepath = asksaveasfilename(
            initialdir=os.path.dirname(self.logfile_entry.get()),
            title="Save scan log to...", parent=self,
            filetypes=(("tag scanner log", "*.log"), ('All', '*')))

        if not filepath:
            return

        filepath = Path(filepath)
        self.app_root.last_load_dir = filepath.parent

        self.logfile_path.set(filepath)

    def destroy(self):
        try:
            self.app_root.tool_windows.pop(self.window_name, None)
        except AttributeError:
            pass
        self.stop_scanning = True
        tk.Toplevel.destroy(self)

    def cancel_scan(self):
        self.stop_scanning = True

    def scan_directory(self):
        if self._scanning:
            return
        try: self.scan_thread.join(1.0)
        except Exception: pass
        self.scan_thread = Thread(target=self._scan_directory, daemon=True)
        self.scan_thread.start()

    def _scan_directory(self):
        self._scanning = True
        try:
            self.scan()
        except Exception:
            print(format_exc())
        self._scanning = False

    def scan(self):
        handler = self.handler
        self.stop_scanning = False

        logpath = path_normalize(self.logfile_path.get())
        dirpath = path_normalize(self.directory_path.get())

        if not is_in_dir(dirpath, self.handler.tagsdir):
            print("Specified directory is not located within the tags directory")
            return

        #this is the string to store the entire debug log
        log_name = "HEK Tag Scanner log"
        debuglog = "\n%s%s%s\n\n" % (
            "-"*30, log_name, "-" * (50-len(log_name)))
        debuglog += "tags directory = %s\nscan directory = %s\n\n" % (
            self.handler.tagsdir, dirpath)
        debuglog += "Broken dependencies are listed below.\n"
        tag_specific_errors = {}

        get_nodes = handler.get_nodes_by_paths
        get_tagref_invalid = handler.get_tagref_invalid

        s_time = time()
        c_time = s_time
        p_int = self.print_interval

        all_tag_paths = {self.listbox_index_to_def_id[int(i)]: [] for i in
                         self.def_ids_listbox.curselection()}
        ext_id_map = handler.ext_id_map
        id_ext_map = handler.id_ext_map

        print("Locating tags...")

        for root, directories, files in os.walk(dirpath):
            root = path_normalize(os.path.join(root, ""))

            rel_root = Path(root).relative_to(self.handler.tagsdir)

            for filename in files:
                filepath = rel_root.joinpath(filename)

                if time() - c_time > p_int:
                    c_time = time()
                    print(' '*4, filepath, sep="")
                    self.app_root.update_idletasks()

                if self.stop_scanning:
                    print('Tag scanning operation cancelled.\n')
                    return

                tag_paths = all_tag_paths.get(
                    ext_id_map.get(os.path.splitext(filename)[-1].lower()))

                if tag_paths is not None:
                    tag_paths.append(filepath)

        # make the debug string by scanning the tags directory
        for def_id in sorted(all_tag_paths.keys()):
            tag_ref_paths = handler.tag_ref_cache.get(def_id)

            self.app_root.update_idletasks()
            print("Scanning '%s' tags..." % id_ext_map[def_id][1:])
            tags_coll = all_tag_paths[def_id]

            # always display the first tag's filepath
            c_time = time() - (p_int + 100)

            for filepath in sorted(tags_coll):
                if self.stop_scanning:
                    print('Tag scanning operation cancelled.\n')
                    break

                if time() - c_time > p_int:
                    c_time = time()
                    print(' '*4, filepath, sep="")
                    self.app_root.update_idletasks()

                tag = self.get_tag(self.handler.tagsdir.joinpath(filepath))
                if tag is None:
                    print("    Could not load '%s'" % filepath)
                    continue

                # find tag specific errors
                self.tag_specific_scan(tag, tag_specific_errors)

                try:
                    if tag_ref_paths is None:
                        # no dependencies for this tag. continue on
                        continue

                    missed = get_nodes(tag_ref_paths, tag.data,
                                       get_tagref_invalid)

                    if not missed:
                        continue

                    debuglog += "\n\n%s\n" % filepath
                    block_name = None

                    for block in missed:
                        if block.NAME != block_name:
                            debuglog += '%s%s\n' % (' '*4, block.NAME)
                            block_name = block.NAME
                        try:
                            ext = '.' + block.tag_class.enum_name
                        except Exception:
                            ext = ''
                        debuglog += '%s%s\n' % (' '*8, block.STEPTREE + ext)

                except Exception:
                    print(format_exc())
                    print("    Could not scan '%s'" % tag.filepath)
                    continue

            if self.stop_scanning:
                break

        if tag_specific_errors:
            debuglog += "\nTag specific errors are listed below.\n"

        for def_id in sorted(tag_specific_errors.keys()):
            debuglog += "\n\n%s specific errors:\n%s" % (
                def_id, tag_specific_errors[def_id])

        print("\nScanning took %s seconds." % int(time() - s_time))
        print("Writing logfile to %s..." % logpath)
        self.app_root.update_idletasks()

        # make and write to the logfile
        try:
            handler.make_log_file(debuglog, logpath)
            try:
                print("Scan completed.\n")
                if self.open_logfile.get():
                    open_in_default_program(logpath)

            except Exception:
                print("Could not open written log.")
            return
        except Exception:
            print("Could not create log. Printing log to console instead.\n\n")
            for line in debuglog.split('\n'):
                try:
                    print(line)
                except Exception:
                    print("<COULD NOT PRINT THIS LINE>")

            print("Scan completed.\n")

    def tag_specific_scan(self, tag, errors):
        assert isinstance(errors, dict)
        cls = tag.def_id
        err = ""

        if cls == "snd!":
            bad_ogg = []
            for pr in tag.data.tagdata.pitch_ranges.STEPTREE:
                for perm in pr.permutations.STEPTREE:
                    if perm.compression.enum_name != "ogg":
                        continue
                    elif perm.ogg_sample_count == 0 and perm.samples.data:
                        bad_ogg.append((pr.name, perm.name))

            if bad_ogg:
                err += ("    Bad PCM buffer size. " +
                        "Fix by recompiling this sound.")
        elif cls == "coll":
            bad_nodes = {}
            tagdata = tag.data.tagdata
            nodes = tagdata.nodes.STEPTREE
            mat_ct = len(tagdata.materials.STEPTREE)
            highest_mat_num = lowest_mat_num = 0

            for i in range(len(nodes)):
                node = nodes[i]
                bad_bsps = {}
                for j in range(len(node.bsps.STEPTREE)):
                    bsp = node.bsps.STEPTREE[j]
                    bad_surfaces = []
                    for k in range(len(bsp.surfaces.STEPTREE)):
                        mat = bsp.surfaces.STEPTREE[k].material
                        if mat > -1 and mat < mat_ct:
                            continue

                        highest_mat_num = max(highest_mat_num, mat)
                        lowest_mat_num  = min(lowest_mat_num, mat)
                        bad_surfaces.append(k)

                    if bad_surfaces:
                        bad_bsps[j] = bad_surfaces

                if bad_bsps:
                    bad_nodes[i] = bad_bsps

            if bad_nodes:
                err += "    Bad collision material numbers.\n"
                if lowest_mat_num > -1:
                    # none of the material numbers are below zero, so
                    # it's possible to fix this by adding more materials.
                    err += (("    Change the material numbers of these " +
                             "surfaces to be <= %s or add %s materials.\n")
                            % (mat_ct - 1, (highest_mat_num + 1) - mat_ct))
                else:
                    err += (("    Change the material numbers of these " +
                             "surfaces to be >= 0 and <= %s\n") % (mat_ct - 1))

                for i in sorted(bad_nodes.keys()):
                    bad_bsps = bad_nodes[i]
                    err += "    %s(node #%s)\n" % (nodes[i].name, i)
                    for j in sorted(bad_bsps.keys()):
                        bad_surfaces = bad_bsps[j]
                        err += "        bsp #%s\n" % j
                        err += "            surfaces = %s\n" % bad_surfaces
                    err += "\n"
                err = err[:-1]

        elif cls == "effe":
            i = 0
            events = tag.data.tagdata.events.STEPTREE
            for i in range(len(events)):
                # tool exceptions if any parts reference a damage effect
                # tag type, but have an empty filepath for the reference
                parts = events[i].parts.STEPTREE
                for j in range(len(parts)):
                    part = parts[j]
                    if (part.type.tag_class.enum_name == "damage_effect" and
                        not part.type.filepath):
                        err += ("     Missing filepath in damage_effect "
                                "reference in part %s of event %s\n." % (j, i))

        if err:
            rel_tag_path = str(tag.filepath.relative_to(self.handler.tagsdir))
            errors[cls] = "%s\n%s\n%s\n" % (
                errors.get(cls, ""), rel_tag_path, err)



#!/usr/bin/env python3
#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from pathlib import Path
from threading import Thread
from time import time
from traceback import format_exc
import os
import threadsafe_tkinter as tk

from binilla.widgets.binilla_widget import BinillaWidget
from binilla.windows.filedialog import askopenfilename, askdirectory
from supyr_struct.util import path_replace, path_split
from mozzarilla import editor_constants as e_c

curr_dir = Path.cwd()

class ConverterBase(BinillaWidget):
    src_ext = "*"
    dst_ext = "*"
    _running_thread = None
    stop_conversion = False

    @property
    def src_exts(self): return (self.src_ext, )

    def __init__(self, app_root, *args, **kwargs):
        self.app_root = app_root
        BinillaWidget.__init__(self, app_root, *args, **kwargs)

    def setup_window(self, *args, **kwargs):
        kwargs.setdefault("title", "%s to %s convertor" %
                          (self.src_ext.capitalize(), self.dst_ext))
        self.title(kwargs.pop("title"))
        self.resizable(0, 0)
        self.update()
        try:
            self.iconbitmap(e_c.MOZZ_ICON_PATH)
        except Exception:
            pass

        # do path_replace to make sure the path works on linux
        tags_dir = Path(path_split(curr_dir, "tags", after=True))
        if self.app_root is not self and hasattr(self.app_root, "tags_dir"):
            tags_dir = getattr(self.app_root, "tags_dir")

        self.tags_dir = tk.StringVar(self, tags_dir)
        self.tag_path = tk.StringVar(self)

        # make the frames
        self.tags_dir_frame = tk.LabelFrame(self, text="Directory of tags")
        self.tag_path_frame = tk.LabelFrame(self, text="Single tag")

        # add the filepath boxes
        self.tags_dir_entry = tk.Entry(
            self.tags_dir_frame, textvariable=self.tags_dir)
        self.tags_dir_entry.config(width=70, state=tk.DISABLED)
        self.tag_path_entry = tk.Entry(
            self.tag_path_frame, textvariable=self.tag_path)
        self.tag_path_entry.config(width=70, state=tk.DISABLED)

        # add the buttons
        self.convert_dir_btn = tk.Button(
            self, text="Convert directory",
            width=15, command=self.convert_dir)
        self.convert_file_btn = tk.Button(
            self, text="Convert tag", width=15,
            command=self.convert_tag)
        self.tags_dir_browse_btn = tk.Button(
            self.tags_dir_frame, text="Browse",
            width=6, command=self.tags_dir_browse)
        self.tag_path_browse_btn = tk.Button(
            self.tag_path_frame, text="Browse",
            width=6, command=self.tag_path_browse)

        if self.app_root and self.app_root is not self:
            self.transient(self.app_root)

    def pack_widgets(self):
        self.tags_dir_entry.pack(expand=True, fill='x', side='left')
        self.tags_dir_browse_btn.pack(fill='x', side='left')
        self.tag_path_entry.pack(expand=True, fill='x', side='left')
        self.tag_path_browse_btn.pack(fill='x', side='left')

        self.tags_dir_frame.pack(expand=True, fill='both')
        self.convert_dir_btn.pack(fill='both', padx=5, pady=5)
        self.tag_path_frame.pack(expand=True, fill='both')
        self.convert_file_btn.pack(fill='both', padx=5, pady=5)

    def destroy(self):
        try:
            self.app_root.tool_windows.pop(self.window_name, None)
        except AttributeError:
            pass
        self.stop_conversion = True

    def apply_style(self, seen=None):
        super(ConverterBase, self).apply_style(seen)
        self.update()
        w, h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry("%sx%s" % (w, h))
        self.minsize(width=w, height=h)

    def tags_dir_browse(self):
        dirpath = askdirectory(initialdir=self.tags_dir.get())
        if dirpath:
            self.tags_dir.set(dirpath)

    def tag_path_browse(self):
        initialdir = self.tag_path.get()
        if not initialdir:
            initialdir = self.tags_dir.get()

        filetypes = (("%s tag" % self.src_ext, "*.%s" % self.src_ext), )
        filetypes += tuple(("%s tag" % ext, "*.%s" % ext)
                          for ext in self.src_exts if ext != self.src_ext)
        filetypes += (('All', '*'), )
        tag_path = askopenfilename(
            initialdir=initialdir, filetypes=filetypes)
        if tag_path:
            self.tag_path.set(tag_path)

    def convert(self, tag_path):
        raise NotImplementedError("Override this method.")

    def lock_ui(self):
        for w in (self.tags_dir_browse_btn, self.tag_path_browse_btn,
                  self.convert_dir_btn, self.convert_file_btn):
            w.config(state=tk.DISABLED)

    def unlock_ui(self):
        for w in (self.tags_dir_browse_btn, self.tag_path_browse_btn,
                  self.convert_dir_btn, self.convert_file_btn):
            w.config(state=tk.NORMAL)

    def thread_wrapper(self, func, *args, **kwargs):
        if self._running_thread is not None:
            return

        try:
            self.lock_ui()
            func(*args, **kwargs)
        finally:
            self.unlock_ui()
            self._running_thread = None

    def convert_tag(self):
        if self._running_thread is None:
            new_thread = Thread(target=self.thread_wrapper,
                                args=(self.do_convert_tag, ))
            new_thread.daemon = True
            new_thread.start()

    def convert_dir(self):
        if self._running_thread is None:
            new_thread = Thread(target=self.thread_wrapper,
                                args=(self.do_convert_dir, ))
            new_thread.daemon = True
            new_thread.start()

    def do_convert_tag(self, tag_path=None, *args, **kwargs):
        start = time()
        part_of_batch = bool(tag_path)
        if tag_path is None:
            tag_path = self.tag_path.get()

        if not os.path.isfile(tag_path):
            return

        print(tag_path)

        try:
            dst_tag = self.convert(tag_path)
            if not part_of_batch and self.stop_conversion:
                print("    Conversion cancelled by user.")
                return

            if dst_tag:
                #print("    Saving to %s" % dst_tag.filepath)
                dst_tag.serialize(temp=False, backup=False, int_test=False)
        except Exception:
            print(format_exc())

        if not part_of_batch:
            print('    Finished. Took %s seconds.\n' % round(time() - start, 1))

    def do_convert_dir(self, *args, **kwargs):
        print("Converting  %s  to  %s" % (self.src_ext, self.dst_ext))
        start = time()
        valid_ext = "." + self.src_ext
        for root, dirs, files in os.walk(self.tags_dir.get()):
            for filename in files:
                if self.stop_conversion:
                    break

                filepath = Path(root, filename)
                if filepath.suffix.lower() == valid_ext:
                    self.do_convert_tag(str(filepath))

            if self.stop_conversion:
                print("    Conversion cancelled by user.")
                break

        print('    Finished. Took %s seconds.\n' % round(time() - start, 1))



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import tkinter as tk

from binilla.widgets.binilla_widget import BinillaWidget
from binilla.widgets.bitmap_display_frame import BitmapDisplayFrame
from binilla.widgets.field_widgets import ContainerFrame, SimpleImageFrame

try:
    import arbytmap
except ImportError:
    arbytmap = None


class FontCharacterDisplayFrame(BitmapDisplayFrame, BinillaWidget):
    def __init__(self, *args, **kwargs):
        BitmapDisplayFrame.__init__(self, *args, **kwargs)
        self.labels_frame = tk.Frame(self, highlightthickness=0)
        self.preview_frame = tk.Frame(self, highlightthickness=0)

        self.font_label0 = tk.Label(self.labels_frame, text="UTF-16 character\t")
        self.font_label1 = tk.Label(self.labels_frame, text="Bitmap preview\t")

        self.preview_label = tk.Label(self.preview_frame, text="")
        self.preview_label.font_type = "font_tag_preview"
        for lbl in (self.font_label0, self.font_label1):
            lbl.config(width=30, anchor='w',
                       disabledforeground=self.text_disabled_color)

        for w in (self.hsb, self.vsb, self.root_canvas):
            w.pack_forget()

        self.image_canvas = tk.Canvas(self.preview_frame, highlightthickness=0)
        padx = self.horizontal_padx
        pady = self.horizontal_pady

        self.image_canvas.config(width=1, height=1)
        self.labels_frame.pack(fill='both', side='left')
        self.preview_frame.pack(fill='both', side='left')
        for w in (self.font_label0, self.font_label1,
                  self.preview_label, self.image_canvas):
            w.pack(fill='x', padx=padx, pady=pady)

        self.apply_style()

    def apply_style(self, seen=None):
        BitmapDisplayFrame.apply_style(self, seen)
        self.image_canvas.config(bg=self.bitmap_canvas_bg_color)


class FontCharacterFrame(SimpleImageFrame):
    display_frame_cls = FontCharacterDisplayFrame

    def reload(self):
        ContainerFrame.reload(self)
        if self.image_frame and self.node:
            node = self.node
            try:
                char_int = node.character
                char = bytes([char_int & 0xFF, char_int >> 8]).decode("utf-16-le")
                width = max(node.bitmap_width, node.character_width, 1)
                height = max(node.bitmap_height, 32)
                self.image_frame().change_textures(self.get_textures())
            except Exception:
                width = height = 1
                char = ""
            self.image_frame().preview_label.config(text=char)
            self.image_frame().image_canvas.config(width=width, height=height)

        self.pose_fields()

    def get_textures(self):
        texture_block = []
        if self.node and self.tag:
            width = min(640, max(self.node.bitmap_width, 1))
            height = min(480, max(self.node.bitmap_height, 1))
            tex_info = dict(width=width, height=height,
                            format=arbytmap.FORMAT_L8)
            pixels_count = width * height
            all_pixels = self.tag.data.tagdata.pixels.data
            pixels = all_pixels[self.node.pixels_offset:
                                self.node.pixels_offset + pixels_count]
            pixels += b'\x00' * (pixels_count - len(pixels))
            texture_block.append([[pixels], tex_info])

        return texture_block



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from mozzarilla import editor_constants as e_c
from binilla.windows.tag_window import TagWindow, ConfigWindow
from supyr_struct.util import is_path_empty

__all__ = ("HaloTagWindow", "HaloConfigWindow", )


class HaloTagWindow(TagWindow):
    def __init__(self, master, tag=None, *args, **kwargs):
        app_root   = kwargs.get('app_root', master)
        is_new_tag = kwargs.get('is_new_tag', self.is_new_tag)
        TagWindow.__init__(self, master, *args, tag=tag, **kwargs)

    def post_toplevel_init(self):
        if not is_path_empty(e_c.MOZZ_ICON_PATH):
            self.iconbitmap_filepath = e_c.MOZZ_ICON_PATH

        TagWindow.post_toplevel_init(self)

    def save(self, **kwargs):
        '''Flushes any lingering changes in the widgets to the tag.'''
        flags = self.app_root.config_file.data.mozzarilla.flags
        tag = self.tag
        kwargs.setdefault("calc_pointers", False)
        if hasattr(self.tag, 'calc_internal_data') and flags.calc_internal_data:
            tag.calc_internal_data()

        try:
            tag.data.blam_header.flags.edited_with_mozz = True
        except Exception:
            pass

        TagWindow.save(self, **kwargs)

    @property
    def use_scenario_names_for_script_names(self):
        try:
            return bool(self.app_root.config_file.data.mozzarilla.\
                        flags.use_scenario_names_in_scripts)
        except Exception:
            return False


class HaloConfigWindow(ConfigWindow):

    def post_toplevel_init(self):
        if not is_path_empty(e_c.MOZZ_ICON_PATH):
            self.iconbitmap_filepath = e_c.MOZZ_ICON_PATH

        ConfigWindow.post_toplevel_init(self)



#!/usr/bin/env python
#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import mozzarilla

long_desc = ""
try:
    long_desc = open("README.MD").read()
except Exception:
    print("Couldn't read readme.")

setup(
    name='mozzarilla',
    description='A variant of Binilla for editing binary structures for '
                'games built with the Blam engine.',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    version='%s.%s.%s' % mozzarilla.__version__,
    url=mozzarilla.__website__,
    project_urls={
        #"Documentation": <Need a string entry here>,
        "Source": mozzarilla.__website__,
        "Funding": "https://liberapay.com/MEK/",
    },
    author=mozzarilla.__author__,
    author_email='MoeMakesStuff@gmail.com',
    license='GPLv3',
    packages=[
        'mozzarilla',
        'mozzarilla.defs',
        'mozzarilla.widgets',
        'mozzarilla.widgets.field_widgets',
        'mozzarilla.windows',
        'mozzarilla.windows.tools',
        'mozzarilla.windows.tag_converters',
        ],
    package_data={
        'mozzarilla': [
            'styles/*.*', '*.[tT][xX][tT]', '*.MD', '*.pyw', '*.ico', '*.png',
            'msg.dat',
            ]
        },
    platforms=["POSIX", "Windows"],
    keywords=["binilla", "binary", "data structure"],
    install_requires=['reclaimer', 'binilla', 'arbytmap', 'supyr_struct'],
    requires=['reclaimer', 'arbytmap', 'binilla'],
    provides=['mozzarilla'],
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        ],
    zip_safe=False,
    )
